##
# File:    SchemaDefUtil.py
# Author:  J. Westbrook
# Date:    31-Mar-2018
# Version: 0.001
#
# Updates:
#   9-Apr-2018 jdw update to provide indices and include remaining schema
#  18-Jun-2018 jdw update to new schema generation protocol
#  22-Jun-2018 jdw change collection attribute specification to dot notation.
#  14-Aug-2018 jdw generalize the primaryIndex to the list of attributes returned by getDocumentKeyAttributeNames()
#  21-Aug-2018 jdw use getHelper() from the configuration class.
#   7-Sep-2018 jdw add method to return a stored JSON schema getJsonSchema()
#  11-Nov-2018 jdw add support for chem_comp_core and bird_chem_comp_core schemas
#   3-Dec-2018 jdw generalize the delivery of document indices
##
"""
 A collection of schema and repo path convenience methods.

"""

__docformat__ = "restructuredtext en"
__author__ = "John Westbrook"
__email__ = "jwest@rcsb.rutgers.edu"
__license__ = "Apache 2.0"


import logging

from rcsb.db.define.SchemaDefAccess import SchemaDefAccess
from rcsb.db.define.SchemaDefBuild import SchemaDefBuild
from rcsb.db.utils.RepoPathUtil import RepoPathUtil
from rcsb.utils.io.MarshalUtil import MarshalUtil

logger = logging.getLogger(__name__)


class SchemaDefUtil(object):
    """ A collection of schema and repository convenience methods.
    """

    def __init__(self, cfgOb=None, numProc=1, fileLimit=None, workPath=None, **kwargs):
        self.__cfgOb = cfgOb
        self.__fileLimit = fileLimit
        self.__numProc = numProc
        self.__workPath = workPath

    def getPathList(self, contentType, inputPathList=None):
        """Convenience method to get the data path list for the input repository content type.

        Args:
            contentType (str): Repository content type (e.g. pdbx, chem_comp, bird, ...)
            inputPathList (list, optional): path list that will be returned if provided.

        Returns:
            list: data file file path list


        """
        inputPathList = inputPathList if inputPathList else []
        rpU = RepoPathUtil(self.__cfgOb, numProc=self.__numProc, fileLimit=self.__fileLimit, workPath=self.__workPath)
        outputPathList = rpU.getRepoPathList(contentType, inputPathList=inputPathList)
        return outputPathList

    def getSchemaInfo(self, contentType, applicationName='ANY', useCache=True, saveSchema=True):
        """Convenience method to return essential schema details for the input repository content type.

        Args:
            contentType (str): Repository content type (e.g. pdbx, bird, chem_comp, ...)
            applicationName (str, optional): Application name for the target schema (e.g. ANY, SQL, ...)
            useCache (bool, optional): Use a cached version of the target schema if available
            saveSchema (bool, optional): If a schema is generated preserve the generated schema

        Returns:
            tuple: SchemaDefAccess(object), target database name, target collection name list, primary index attribute list


        """
        sd = None
        dbName = None
        collectionNameList = []
        docIndexD = {}
        mU = MarshalUtil(workPath=self.__workPath)
        try:
            optName = 'SCHEMA_DEF_LOCATOR_%s' % applicationName.upper()
            schemaLocator = self.__cfgOb.getPath(optName, sectionName=contentType)
            schemaDef = mU.doImport(schemaLocator, format="json")
            if schemaDef and useCache:
                logger.debug("Using cached schema definition for %s application %s" % (contentType, applicationName))
                sd = SchemaDefAccess(schemaDef)
            else:
                schemaDef = self.__buildSchema(contentType, applicationName)
                if schemaDef and saveSchema:
                    optName = 'SCHEMA_DEF_LOCATOR_%s' % applicationName.upper()
                    schemaLocator = self.__cfgOb.getPath(optName, sectionName=contentType)
                    mU.doExport(schemaLocator, schemaDef, format="json")
                sd = SchemaDefAccess(schemaDef)

            if sd:
                dbName = sd.getDatabaseName()
                collectionNameList = sd.getContentTypeCollections(contentType)
                logger.debug("Schema %s database name %s collections %r" % (contentType, dbName, collectionNameList))
                for collectionName in collectionNameList:
                    docIndexD[collectionName] = sd.getDocumentIndices(collectionName)

        except Exception as e:
            logger.exception("Retreiving schema %s for %s failing with %s" % (contentType, applicationName, str(e)))

        return sd, dbName, collectionNameList, docIndexD

    def getJsonSchema(self, collectionName, level='full'):
        """Return JSON schema (w/ BSON types) for the input collection and level.and

           Currenting using configuration information -

            pdbx_core_entity_v5_0_2]
                SCHEMA_NAME=pdbx_core
                JSON_SCHEMA_FULL_LOCATOR=...
                JSON_SCHEMA_MIN_LOCATOR=...

        Args:
            collectionName (str): Collection name in document store
            level (str, optional): Completeness of the schema (e.g. min or full)

        Returns:
            dict: Schema object

        """
        sObj = None
        if level.lower() == 'full':
            schemaLocator = self.__cfgOb.getPath('BSON_SCHEMA_FULL_LOCATOR', sectionName=collectionName)
        elif level.lower() in ['min', 'minimum']:
            schemaLocator = self.__cfgOb.getPath('BSON_SCHEMA_MIN_LOCATOR', sectionName=collectionName)
        else:
            logger.error("Unsupported schema level %s %r" % (collectionName, level))
            schemaLocator = None

        if schemaLocator:
            #
            mU = MarshalUtil(workPath=self.__workPath)
            sObj = mU.doImport(schemaLocator, format="json")
        else:
            logger.error("Failed to read schema for %s %r" % (collectionName, level))
        return sObj

    def __buildSchema(self, contentType, applicationName):
        """Internal method to create a schema definition from dictionary and supported metadata.


        Args:
            contentType (str): Repository content type (e.g. pdbx, bird, chem_comp, ...)
            applicationName (str, optional): Application name for the target schema (e.g. ANY, SQL, ...)

        Returns:
            dict: schema definition


        """
        try:
            #
            logger.debug("Building schema definition for %s application %s" % (contentType, applicationName))
            smb = SchemaDefBuild(contentType, self.__cfgOb.getConfigPath(), mockTopPath=self.__cfgOb.getMockTopPath())
            #
            sD = smb.build(applicationName=applicationName, schemaType='rcsb')
            logger.info("Schema %s dictionary category length %d" % (contentType, len(sD['SCHEMA_DICT'])))
            return sD

        except Exception as e:
            logger.exception("Building schema %s failing with %s" % (contentType, str(e)))

        return {}



##
# File:    CrateDbLoader.py
# Author:  J. Westbrook
# Date:    1-Apr-2018
#
#  Loader variant to support stripped down support for Crate DB.
#
# Updates:
#
##
"""
Generic mapper of PDBx/mmCIF instance data to SQL loadable data files based on external
schema definition defined in class SchemaDefBase().

"""
__docformat__ = "restructuredtext en"
__author__ = "John Westbrook"
__email__ = "jwest@rcsb.rutgers.edu"
__license__ = "Apache 2.0"


import logging
import time

from rcsb.db.crate.CrateDbUtil import CrateDbQuery
from rcsb.db.processors.SchemaDefDataPrep import SchemaDefDataPrep
from rcsb.db.sql.SqlGen import SqlGenAdmin

#
try:
    from mmcif.io.IoAdapterCore import IoAdapterCore as IoAdapter
except Exception:
    from mmcif.io.IoAdapterPy import IoAdapterPy as IoAdapter

logger = logging.getLogger(__name__)


class CrateDbLoader(object):

    """ Map PDBx/mmCIF instance data to SQL loadable data using external schema definition.
    """

    def __init__(self, schemaDefObj, ioObj=None, dbCon=None, workPath='.', cleanUp=False, warnings='default', verbose=True):
        self.__verbose = verbose
        self.__debug = False
        self.__sD = schemaDefObj
        self.__ioObj = ioObj
        #
        self.__dbCon = dbCon
        self.__workingPath = workPath
        self.__pathList = []
        self.__cleanUp = cleanUp
        #
        self.__sdp = SchemaDefDataPrep(schemaDefAccessObj=schemaDefObj, ioObj=IoAdapter(), verbose=True)

    def load(self, inputPathList=None, containerList=None, loadType='batch-file', deleteOpt=None, tableIdSkipD=None):
        """ Load data for each table defined in the current schema definition object.
            Data are extracted from the input path or container list.

            Data source options:

              inputPathList = [<full path of target input file>, ....]

            or

              containerList = [ data container, ...]


            loadType  =  ['crate-insert' | 'crate-insert-many']
            deleteOpt = 'selected' | 'all'

            tableIdSkipD - searchable container with tableIds to be skipped on loading -

            Loading is performed using the current database server connection.

            Intermediate data files for 'batch-file' loading are created in the current working path.

            Returns True for success or False otherwise.

        """
        tableIdSkipD = tableIdSkipD if tableIdSkipD is not None else {}
        if inputPathList is not None:
            tableDataDict, containerNameList = self.__sdp.fetch(inputPathList)
        elif containerList is not None:
            tableDataDict, containerNameList = self.__sdp.process(containerList)
        #

        #
        if loadType in ['crate-insert', 'crate-insert-many']:
            sqlMode = 'single'
            if loadType in ['crate-insert-many']:
                sqlMode = 'many'
            for tableId, rowList in tableDataDict.items():
                if tableId in tableIdSkipD:
                    continue
                if deleteOpt in ['all', 'selected'] or len(rowList) > 0:
                    self.__crateInsertImport(tableId, rowList=rowList, containerNameList=containerNameList, deleteOpt=deleteOpt, sqlMode=sqlMode)
            return True
        else:
            pass

        return False

    def __crateInsertImport(self, tableId, rowList=None, containerNameList=None, deleteOpt='selected', sqlMode='many', refresh=True):
        """ Load the input table using sql crate templated inserts of the input rowlist of dictionaries (i.e. d[attributeId]=value).

            The containerNameList corresponding to the data within loadable data in rowList can be provided
            if 'selected' deletions are to performed prior to the the batch data inserts.

            deleteOpt = ['selected','all'] where 'selected' deletes rows corresponding to the input container
                        list before insert.   The 'all' options truncates the table prior to insert.

                        Deletions are performed in the absence of loadable data.

        """
        startTime = time.time()
        sqlRefresh = None
        crQ = CrateDbQuery(dbcon=self.__dbCon, verbose=self.__verbose)
        sqlGen = SqlGenAdmin(self.__verbose)
        #
        databaseName = self.__sD.getVersionedDatabaseName()
        tableDefObj = self.__sD.getSchemaObject(tableId)
        tableName = tableDefObj.getName()
        tableAttributeIdList = tableDefObj.getAttributeIdList()
        tableAttributeNameList = tableDefObj.getAttributeNameList()
        #
        sqlDeleteList = None
        if deleteOpt in ['selected', 'delete'] and containerNameList is not None:
            deleteAttributeName = tableDefObj.getDeleteAttributeName()
            sqlDeleteList = sqlGen.deleteFromListSQL(databaseName, tableName, deleteAttributeName, containerNameList, chunkSize=10)
            logger.debug("Delete SQL for %s : %r" % (tableId, sqlDeleteList))
        elif deleteOpt in ['all', 'truncate']:
            sqlDeleteList = [sqlGen.truncateTableSQL(databaseName, tableName)]
        #
        logger.debug("Deleting from table %s length %d" % (tableName, len(containerNameList)))
        crQ.sqlCommandList(sqlDeleteList)
        logger.debug("Delete commands %s" % sqlDeleteList)
        if len(rowList) == 0:
            return True
        if refresh:
            sqlRefresh = sqlGen.refreshTableSQLCrate(databaseName, tableName)
            crQ.sqlCommand(sqlRefresh)
        #
        logger.info("Insert begins for table %s with row length %d" % (tableName, len(rowList)))
        sqlInsertList = []
        tupL = list(zip(tableAttributeIdList, tableAttributeNameList))
        if sqlMode == 'many':
            aList = []
            for id, nm in tupL:
                aList.append(nm)
            #
            vLists = []
            for row in rowList:
                vList = []
                for id, nm in tupL:
                    if row[id] and row[id] != r'\N':
                        vList.append(row[id])
                    else:
                        vList.append(None)
                vLists.append(vList)
            #
            lenT = len(vLists)
            lenR = crQ.sqlTemplateCommandMany(sqlTemplate=sqlGen.insertTemplateSQLCrate(databaseName, tableName, aList), valueLists=vLists)
            ret = (lenR == len(vLists))
        else:
            aList = []
            for id, nm in tupL:
                aList.append(nm)
            #
            for row in rowList:
                vList = []
                for id, nm in tupL:
                    if row[id] and row[id] != r'\N':
                        vList.append(row[id])
                    else:
                        vList.append(None)
                sqlInsertList.append((sqlGen.insertTemplateSQLCrate(databaseName, tableName, aList), vList))
            #
            lenT = len(sqlInsertList)
            lenR = crQ.sqlTemplateCommandList(sqlInsertList)
            ret = (lenR == lenT)
        if refresh:
            sqlRefresh = sqlGen.refreshTableSQLCrate(databaseName, tableName)
            crQ.sqlCommand(sqlRefresh)
        #
        endTime = time.time()
        if (ret):
            logger.info("Insert succeeds for table %s %d of %d rows at %s (%.3f seconds)" %
                        (tableName, lenR, lenT, time.strftime("%Y %m %d %H:%M:%S", time.localtime()), endTime - startTime))
        else:
            logger.info("Insert fails for table %s %d of %d rows at %s (%.3f seconds)" %
                        (tableName, lenR, lenT, time.strftime("%Y %m %d %H:%M:%S", time.localtime()), endTime - startTime))

        return ret


if __name__ == '__main__':
    pass



##
# File: ChemRefEtlWorker.py
# Date: 2-Jul-2018  jdw
#
# ETL utilities for processing chemical reference data and related data integration.
#
# Updates:
#  9-Dec-2018  jdw add validation methods
#
##
__docformat__ = "restructuredtext en"
__author__ = "John Westbrook"
__email__ = "jwest@rcsb.rutgers.edu"
__license__ = "Apache 2.0"

import logging

from jsonschema import Draft4Validator, FormatChecker

from rcsb.db.define.SchemaDefBuild import SchemaDefBuild
from rcsb.db.mongo.Connection import Connection
from rcsb.db.mongo.DocumentLoader import DocumentLoader
from rcsb.db.mongo.MongoDbUtil import MongoDbUtil
from rcsb.db.processors.DataExchangeStatus import DataExchangeStatus
from rcsb.db.utils.SchemaDefUtil import SchemaDefUtil
from rcsb.utils.chemref.ChemRefDataPrep import ChemRefDataPrep
from rcsb.utils.io.IoUtil import IoUtil

logger = logging.getLogger(__name__)


class ChemRefEtlWorker(object):
    """ Prepare and load repository holdings and repository update data.
    """

    def __init__(self, cfgOb, workPath=None, numProc=2, chunkSize=10, readBackCheck=False, documentLimit=None, mockTopPath=None, verbose=False):
        self.__cfgOb = cfgOb
        self.__workPath = workPath
        self.__readBackCheck = readBackCheck
        self.__numProc = numProc
        self.__chunkSize = chunkSize
        self.__documentLimit = documentLimit
        self.__mockTopPath = mockTopPath
        #
        self.__resourceName = "MONGO_DB"
        self.__filterType = "assign-dates"
        self.__verbose = verbose
        self.__statusList = []
        self.__schU = SchemaDefUtil(cfgOb=self.__cfgOb, numProc=self.__numProc, fileLimit=self.__documentLimit, workPath=self.__workPath)
        #

    def getChemCompAccesionMapping(self, extResource, **kwargs):
        """ Get the accession code mapping between chemical component identifiers and identifier(s) for the
            input external resource.

        Args:
            resourceName (str):  resource name (e.g. DrugBank, CCDC)
            **kwargs: unused

        Returns:
            dict: {dbExtId: True, dbExtId: True, ...  }

        """
        idD = {}
        try:
            docList = []
            with Connection(cfgOb=self.__cfgOb, resourceName=self.__resourceName) as client:
                mg = MongoDbUtil(client)
                if mg.collectionExists("chem_comp_v5", "chem_comp_core_v5_0_2"):
                    logger.info("Document count is %d" % mg.count("chem_comp_v5", "chem_comp_core_v5_0_2"))
                    qD = {'rcsb_chem_comp_related.resource_name': extResource}
                    selectL = ['rcsb_chem_comp_related', '__comp_id']
                    tL = mg.fetch("chem_comp_v5", "chem_comp_core_v5_0_2", selectL, queryD=qD)
                    logger.info("CC mapping count %d" % len(tL))
                    docList.extend(tL)

                if mg.collectionExists("chem_comp_v5", "bird_chem_comp_core_v5_0_2"):
                    qD = {'rcsb_chem_comp_related.resource_name': extResource}
                    selectL = ['rcsb_chem_comp_related', '__prd_id']
                    tL = mg.fetch("chem_comp_v5", "bird_chem_comp_core_v5_0_2", selectL, queryD=qD)
                    logger.info("BIRD mapping count %d" % len(tL))
                    docList.extend(tL)
                #
                for doc in docList:
                    dL = doc['rcsb_chem_comp_related'] if 'rcsb_chem_comp_related' in doc else []
                    for d in dL:
                        if d['resource_name'] == extResource and 'resource_accession_code' in d:
                            idD[d['resource_accession_code']] = True
        except Exception as e:
            logger.exception("Failing with %s" % str(e))
        return idD

    def __updateStatus(self, updateId, databaseName, collectionName, status, startTimestamp):
        try:
            sFlag = 'Y' if status else 'N'
            desp = DataExchangeStatus()
            desp.setStartTime(tS=startTimestamp)
            desp.setObject(databaseName, collectionName)
            desp.setStatus(updateId=updateId, successFlag=sFlag)
            desp.setEndTime()
            self.__statusList.append(desp.getStatus())
            return True
        except Exception as e:
            logger.exception("Failing with %s" % str(e))
        return False

    def load(self, updateId, extResource, loadType='full'):
        """ Load chemical reference integrated data -

        Relevant configuration options:

        site_info:
            DRUGBANK_MAPPING_LOCATOR: DrugBank/drugbank_pdb_mapping.json
            DRUGBANK_DATA_LOCATOR: DrugBank/full_database.xml.gz

        drugbank_core:
            DATABASE_NAME: chem_comp_core
            DATABASE_VERSION_STRING: v5
            COLLECTION_DRUGBANK_CORE: drugbank_core
            COLLECTION_VERSION_STRING: v0_1
            SCHEMA_DEF_LOCATOR_SQL: schema/schema_def-drugbank_core-SQL.json
            SCHEMA_DEF_LOCATOR_ANY: schema/schema_def-drugbank_core-ANY.json
            INSTANCE_DATA_TYPE_INFO_LOCATOR: data_type_info/scan-drugbank_core-type-map.json

        """
        try:
            self.__statusList
            desp = DataExchangeStatus()
            statusStartTimestamp = desp.setStartTime()
            #
            crdp = ChemRefDataPrep(self.__cfgOb)
            idD = self.getChemCompAccesionMapping(extResource)
            dList = crdp.getDocuments(extResource, idD)
            #
            logger.info("Resource %r mapped document length %d" % (extResource, len(dList)))
            logger.debug("Objects %r" % dList[:2])

            sectionName = 'drugbank_core'
            #
            dl = DocumentLoader(self.__cfgOb, self.__resourceName, numProc=self.__numProc, chunkSize=self.__chunkSize,
                                documentLimit=self.__documentLimit, verbose=self.__verbose, readBackCheck=self.__readBackCheck)
            #
            databaseName = self.__cfgOb.get('DATABASE_NAME', sectionName=sectionName) + '_' + self.__cfgOb.get('DATABASE_VERSION_STRING', sectionName=sectionName)
            collectionVersion = self.__cfgOb.get('COLLECTION_VERSION_STRING', sectionName=sectionName)
            collectionName = self.__cfgOb.get('COLLECTION_DRUGBANK_CORE', sectionName=sectionName) + '_' + collectionVersion
            ok = dl.load(databaseName, collectionName, loadType=loadType, documentList=dList,
                         indexAttributeList=['update_id'], keyNames=None)

            self.__updateStatus(updateId, databaseName, collectionName, ok, statusStartTimestamp)

            return True
        except Exception as e:
            logger.exception("Failing with %s" % str(e))
        return False

    def validate(self, extResource, schemaPath=None, dataPath=None):
        eCount = 0
        if extResource == "DrugBank":
            schemaName = 'drugbank_core'
            collectionNames = ['drugbank_core_v0_1']
            crdp = ChemRefDataPrep(self.__cfgOb)
            idD = self.getChemCompAccesionMapping(extResource)
            dList = crdp.getDocuments(extResource, idD)
            if dataPath:
                ioU = IoUtil()
                ioU.serialize(dataPath, dList[0], format='json', indent=3)
            eCount = self.__validate(schemaPath, schemaName, collectionNames, dList, enforceOpts="mandatoryKeys|mandatoryAttributes|bounds|enums")

        return eCount

    def __validate(self, schemaPath, schemaName, collectionNames, dList, enforceOpts="mandatoryKeys|mandatoryAttributes|bounds|enums"):

        eCount = 0
        for collectionName in collectionNames:
            cD = self.__buildJsonSchema(schemaPath, schemaName, collectionName, enforceOpts=enforceOpts)
            # Raises exceptions for schema compliance.
            Draft4Validator.check_schema(cD)
            #
            v = Draft4Validator(cD, format_checker=FormatChecker())
            for ii, d in enumerate(dList):
                logger.debug("Schema %s collection %s document %d" % (schemaName, collectionName, ii))
                try:
                    cCount = 0
                    for error in sorted(v.iter_errors(d), key=str):
                        logger.info("schema %s collection %s path %s error: %s" % (schemaName, collectionName, error.path, error.message))
                        logger.info(">>> failing object is %r" % d)
                        eCount += 1
                        cCount += 1
                    #
                    logger.debug("schema %s collection %s count %d" % (schemaName, collectionName, cCount))
                except Exception as e:
                    logger.exception("Validation error %s" % str(e))

        return eCount

    def __buildJsonSchema(self, schemaPath, schemaName, collectionName, enforceOpts="mandatoryKeys|mandatoryAttributes|bounds|enums"):
        try:
            smb = SchemaDefBuild(schemaName, self.__cfgOb.getConfigPath(), mockTopPath=self.__mockTopPath)
            cD = smb.build(collectionName, applicationName='json', schemaType='json', enforceOpts=enforceOpts)
            #
            logger.debug("Schema dictionary category length %d" % len(cD['properties']))
            #
            if schemaPath:
                ioU = IoUtil()
                ioU.serialize(schemaPath, cD, format='json', indent=3)
            return cD

        except Exception as e:
            logger.exception("Failing with %s" % str(e))

    def getLoadStatus(self):
        return self.__statusList



##
# File:    TimeUtil.py
# Author:  J. Westbrook
# Date:    23-Jun-2018
# Version: 0.001
#
# Updates:
##
"""
Convenience utilities to manipulate timestamps.

"""

__docformat__ = "restructuredtext en"
__author__ = "John Westbrook"
__email__ = "jwest@rcsb.rutgers.edu"
__license__ = "Apache 2.0"

import datetime
import logging

import dateutil.parser
from dateutil.tz import tzlocal

import pytz

logger = logging.getLogger(__name__)


class TimeUtil(object):

    def __init__(self, **kwargs):
        pass

    def getTimestamp(self, useUtc=True):
        """Return an a pseudo ISO 8601 format timestamp string (2018-07-11 12:33:22.874957+00:00) including timezone details.

        Args:
            useUtc (bool, optional): Use UTC time reference

        Returns:
            str: ISO 8601 format timestamp string
        """
        dt = datetime.datetime.utcnow().replace(tzinfo=pytz.utc) if useUtc else datetime.datetime.now().replace(tzinfo=tzlocal())
        return dt.isoformat(' ')

    def getWeekSignature(self, yyyy, mm, dd):
        """Return week in year signature (e.g. 2018_21) for the input date (year, month and day).

        Args:
            yyyy (int): year
            mm (int): month in year (1-12)
            dd (int): day in month (1-##)

        Returns:
            str: week in year signature (<yyyy>_<week_number>)
        """
        return datetime.date(yyyy, mm, dd).strftime("%Y_%V")

    def getCurrentWeekSignature(self):
        """Return the curren tweek in year signature (e.g. 2018_21).

        Returns:
            str: week in year signature (<yyyy>_<week_number>)
        """
        dt = datetime.date.today()
        return dt.strftime("%Y_%V")

    def getDateTimeObj(self, tS):
        """Return a datetime object corresponding to the input timestamp string.

        Args:
            tS (str): timestamp string (e.g. 2018-07-11 12:33:22.874957+00:00)

        Returns:
            object: datetime object
        """
        try:
            return dateutil.parser.parse(tS)
        except Exception as e:
            logger.exception("Failing with %r" % str(e))
        return None



##
# File:    DataTypeApplicationInfo.py
# Author:  J. Westbrook
# Date:    19-Apr-2018
# Version: 0.001 Initial version
#
# Updates:
#  22-May-2018 jdw standardize data names and fix alignment in default data sections
#  23-May-2018 jdw change assumptions for update method and add tests.
#   7-Jun-2018 jdw rename and rescope.
#  15-Aug-2018 jdw add mapping for JSON types based on generic 'ANY' typing.
#  29-Sep-2018 jdw make JSON date and datetime type explicit in JSON
#  12-Oct-2018 jdw unsuppress datetime mapping
#
##
"""
Manage mapping of default application data type to dictionary data types.

"""
__docformat__ = "restructuredtext en"
__author__ = "John Westbrook"
__email__ = "jwest@rcsb.rutgers.edu"
__license__ = "Apache 2.0"

import copy
import logging

from mmcif.api.DataCategory import DataCategory
from mmcif.api.PdbxContainers import DataContainer

from rcsb.utils.io.MarshalUtil import MarshalUtil

logger = logging.getLogger(__name__)


class DataTypeApplicationInfo(object):

    cifTypes = [
        "code", "ucode", "line", "uline", "text", "int", "float", "name",
        "idname", "any", "yyyy-mm-dd", "uchar3", "uchar1", "symop", "atcode",
        "yyyy-mm-dd:hh:mm", "fax", "phone", "email", "code30", "float-range",
        "operation_expression", "yyyy-mm-dd:hh:mm-flex", "ec-type", "ucode-alphanum-csv",
        "int-range", "point_symmetry", "id_list", "4x3_matrix", "non_negative_int", "positive_int", "emd_id",
        "pdb_id", "point_group", "point_group_helical", "boolean", "author", "orcid_id", "symmetry_operation", ""
    ]
    # These are generic types -
    appTypes = [
        "char", "char", "char", "char", "char", "int", "float", "char", "char",
        "text", "date", "char", "char", "char", "char", "datetime", "char",
        "char", "char", "char", "char", "char", "datetime", "char", "char",
        "char", "char", "char", "char", "int", "int", "char", "char", "char", "char", "char", "char", "char", "char", ""
    ]

    defaultWidths = [
        "10", "10", "80", "80", "200", "10", "10", "80", "80", "255", "15", "4",
        "2", "10", "6", "20", "25", "25", "80", "30", "30", "30", "20", "10",
        "25", "20", "80", "100", "10", "10", "10", "15", "20", "20", "20", "5", "80", "20", "80", ""
    ]

    defaultPrecisions = [
        "0", "0", "0", "0", "0", "0", "6", "0", "0", "0", "0", "0", "0", "0",
        "0", "0", "0", "0", "0", "0", "0", "0", "0", "0",
        "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", ""
    ]

    #
    def __init__(self, locator=None, applicationName='ANY', workPath=None, **kwargs):
        self.__workPath = workPath
        self.__locator = locator
        self.__applicationName = applicationName
        self.__maxCharWidth = 32768
        self.__dtmD = self.__setup(self.__locator, self.__applicationName)

    def __setup(self, locator, applicationName):
        appName = 'ANY' if applicationName in ['JSON', 'BSON'] else applicationName
        if locator:
            dtmD = self.readDefaultDataTypeMap(locator, applicationName=appName)
        else:
            logger.debug(">>>> Falling back to default type mapping. ")
            dtmD = self.getDefaultDataTypeMap(applicationName=appName)
        # for JSON - transform the generic 'ANY' data types -
        if applicationName == 'JSON':
            for cifType, tD in dtmD.items():
                if tD['application_name'] == 'ANY':
                    if tD['app_type_code'] in ['char', 'text']:
                        tD['app_type_code'] = 'string'
                    elif tD['app_type_code'] in ['date', 'datetime']:
                        tD['app_type_code'] = tD['app_type_code']
                    elif tD['app_type_code'] in ['float']:
                        tD['app_type_code'] = 'number'
                    elif tD['app_type_code'] in ['int']:
                        tD['app_type_code'] = 'integer'
                tD['application_name'] = 'JSON'
        elif applicationName == 'BSON':
            for cifType, tD in dtmD.items():
                if tD['application_name'] == 'ANY':
                    if tD['app_type_code'] in ['char', 'text']:
                        tD['app_type_code'] = 'string'
                    elif tD['app_type_code'] in ['float', 'double']:
                        tD['app_type_code'] = 'double'
                    elif tD['app_type_code'] in ['date', 'datetime']:
                        tD['app_type_code'] = 'date'
                    elif tD['app_type_code'] in ['int', 'integer']:
                        tD['app_type_code'] = 'int'
                tD['application_name'] = 'BSON'
        #
        return dtmD

    def getDefaultDataTypeMap(self, applicationName='ANY'):
        try:
            mapD = {}
            for (cifType, simpleType, defWidth, defPrecision) in zip(DataTypeApplicationInfo.cifTypes,
                                                                     DataTypeApplicationInfo.appTypes,
                                                                     DataTypeApplicationInfo.defaultWidths,
                                                                     DataTypeApplicationInfo.defaultPrecisions):
                if self.__isNull(cifType):
                    continue
                mapD[cifType] = {
                    'app_type_code': simpleType,
                    'app_precision_default': defPrecision,
                    'app_width_default': defWidth,
                    'type_code': cifType,
                    'application_name': applicationName}
            return mapD
        except Exception as e:
            logger.exception("Failing with %s" % str(e))
        return {}

    def updateCharType(self, isKey, appType, dataWidth, defaultWidth, bufferPercent=30.0, minWidth=10):
        """ Skeleton implementation needs to add bracket typing around [80, 256, 1024, 2048, ... ]
        """
        retDataWidth = defaultWidth
        retDataType = appType
        iWidth = dataWidth + int(bufferPercent * 0.01 * dataWidth)
        retDataWidth = iWidth if iWidth > minWidth else minWidth
        if self.__applicationName.upper() in ['SQL', 'MYSQL', 'COCKROACH', 'CRATE']:

            if appType.upper() in ['CHAR', 'VARCHAR'] and retDataWidth > self.__maxCharWidth:
                retDataType = 'TEXT'
            if appType.upper() in ['TEXT'] and isKey:
                retDataType = 'CHAR'
        else:
            pass
        return (retDataType, retDataWidth)

    def hasType(self, cifType):
        """
        """
        try:
            return cifType in self.__dtmD
        except Exception:
            return False

    def getAppTypeName(self, cifType, default=None):
        """
        """
        try:
            return self.__dtmD[cifType]['app_type_code']
        except Exception:
            return default

    def getAppTypeDefaultWidth(self, cifType, default=None):
        """
        """
        try:
            return self.__dtmD[cifType]['app_width_default']
        except Exception:
            return default

    def getAppTypeDefaultPrecision(self, cifType, default=None):
        """
        """
        try:
            return self.__dtmD[cifType]['app_precision_default']
        except Exception:
            return default

    def writeDefaultDataTypeMap(self, outPath, applicationName='ANY'):
        """ Write data file containing application default dictionary to application data type mapping

                  data_rcsb_data_type_map
                    loop_
                    _pdbx_data_type_application_map.application_name
                    _pdbx_data_type_application_map.type_code
                    _pdbx_data_type_application_map.app_type_code
                    _pdbx_data_type_application_map.app_precision_default
                    _pdbx_data_type_application_map.app_width_default
                    # .... type mapping data ...
        """
        try:
            #
            containerList = []
            curContainer = DataContainer("rcsb_data_type_map")
            aCat = DataCategory("pdbx_data_type_application_map")
            aCat.appendAttribute("application_name")
            aCat.appendAttribute("type_code")
            aCat.appendAttribute("app_type_code")
            aCat.appendAttribute("app_width_default")
            aCat.appendAttribute("app_precision_default")
            for (cifType, simpleType, defWidth, defPrecision) in zip(DataTypeApplicationInfo.cifTypes,
                                                                     DataTypeApplicationInfo.appTypes,
                                                                     DataTypeApplicationInfo.defaultWidths,
                                                                     DataTypeApplicationInfo.defaultPrecisions):
                if self.__isNull(cifType):
                    continue
                aCat.append([applicationName, cifType, simpleType, defWidth, defPrecision])
            curContainer.append(aCat)
            containerList.append(curContainer)
            #
            mU = MarshalUtil(workPath=self.__workPath)
            ok = mU.doExport(outPath, containerList, format="mmcif", enforceAscii=True, useCharRefs=True, raiseExceptions=True)

            return ok
        except Exception as e:
            logger.exception("Failing with %s" % str(e))
        return False

    def updateDefaultDataTypeMap(self, filePath, mapD, applicationName='ANY'):
        """ Update data file containing application default data type mapping with any
            updates from the input type mapping dictionary

           mapD['cif_type_code'] -> ['application_name', 'app_type_code', 'app_precision_default', 'app_width_default', 'type_code']

                  data_rcsb_data_type_map
                    loop_
                    _pdbx_data_type_application_map.application_name
                    _pdbx_data_type_application_map.type_code
                    _pdbx_data_type_application_map.app_type_code
                    _pdbx_data_type_application_map.app_precision_default
                    _pdbx_data_type_application_map.app_width_default
                    # .... type mapping data ...
        """
        try:
            #
            mD = copy.deepcopy(mapD)
            mU = MarshalUtil(workPath=self.__workPath)
            containerList = mU.doImport(filePath, format="mmcif", enforceAscii=True, useCharRefs=True, raiseExceptions=True)
            for container in containerList:
                if container.getName() == 'rcsb_data_type_map':
                    catObj = container.getObj('pdbx_data_type_application_map')
                    rIL = []
                    for ii in range(catObj.getRowCount()):
                        d = catObj.getRowAttributeDict(ii)
                        if d['application_name'] == applicationName:
                            rIL.append(ii)
                            mD[d['type_code']] = {k: d[k] for k in ['application_name', 'app_type_code', 'app_precision_default', 'app_width_default', 'type_code']}
                            continue
                    ok = catObj.removeRows(rIL)
                    atNameL = catObj.getAttributeList()
                    for ky in mapD:
                        r = [mapD[ky][atN] for atN in atNameL]
                        catObj.append(r)
            #
            # Write updated data file
            mU = MarshalUtil(workPath=self.__workPath)
            ok = mU.doExport(filePath, containerList, format="mmcif", enforceAscii=True, useCharRefs=True, raiseExceptions=True)

            return ok
        except Exception as e:
            logger.exception("Failing with %s" % str(e))
        return False

    def readDefaultDataTypeMap(self, locator, applicationName='ANY'):
        """ Read data file containing application default data type mapping

                  data_rcsb_data_type_map
                    loop_
                    _pdbx_data_type_application_map.application_name
                    _pdbx_data_type_application_map.type_code
                    _pdbx_data_type_application_map.app_type_code
                    _pdbx_data_type_application_map.app_precision_default
                    _pdbx_data_type_application_map.app_width_default
                    # .... type mapping data ...

            Return (dict):  map[cifType] -> appType, width, precision
                        mapD['cif_type_code'] -> ['application_name', 'app_type_code', 'app_precision_default', 'app_width_default', 'type_code']
        """
        try:
            #
            mapD = {}
            mU = MarshalUtil(workPath=self.__workPath)
            containerList = mU.doImport(locator, format="mmcif", enforceAscii=True, useCharRefs=True, raiseExceptions=True)

            for container in containerList:
                if container.getName() == 'rcsb_data_type_map':
                    catObj = container.getObj('pdbx_data_type_application_map')
                    for ii in range(catObj.getRowCount()):
                        d = catObj.getRowAttributeDict(ii)
                        if d['application_name'] == applicationName:
                            mapD[d['type_code']] = {k: d[k] for k in ['app_type_code', 'application_name', 'type_code']}
                            mapD[d['type_code']].update({k: int(d[k]) for k in ['app_precision_default', 'app_width_default']})
            return mapD
        except Exception as e:
            logger.exception("Failing with %s" % str(e))
        return {}

    def __isNull(self, value):
        if not value:
            return True
        if (len(value) == 0) or (value == '?') or (value == '.'):
            return True
        return False



##
# File:    DataTypeInstanceInfo.py
# Author:  J. Westbrook
# Date:    7-Jun-2018
# Version: 0.001 Initial version
#
# Updates:
#      15-Jun-2018 jdw cleanup exception handling
#      18-Jun-2018 jdw turn off distracting warning messages by default
#       6-Oct-2018 jdw make the methods in this module an innocuous pass-thru
#                      if no instance data is available.
#
##
"""
Manage data type details extracted by scanning example data sets.

"""
__docformat__ = "restructuredtext en"
__author__ = "John Westbrook"
__email__ = "jwest@rcsb.rutgers.edu"
__license__ = "Apache 2.0"

import logging

from rcsb.utils.io.IoUtil import IoUtil

logger = logging.getLogger(__name__)


class DataTypeInstanceInfo(object):

    def __init__(self, filePath, **kwargs):
        self.__filePath = filePath
        # Turn off warnings for missing values
        self.__verbose = kwargs.get("verbose", False)
        self.__tD = {}
        self.__ioU = IoUtil()
        self.__byPassMode = not self.__setup(self.__filePath)

    def __setup(self, filePath):
        """
           Read the output serialized by ScanRepoUtil() -
           tD[category] -> d[atName]->{minWidth: , maxWidth:, minPrec:, maxPrec: , count}
        """
        if self.__ioU.exists(filePath):
            self.__tD = self.__ioU.deserialize(filePath, format='json')
            return len(self.__tD) > 0
        else:
            return False

    def exists(self, catName, atName=None):
        if self.__byPassMode:
            return True
        try:
            if atName:
                return atName in self.__tD[catName]
            else:
                return catName in self.__tD
            return True
        except Exception:
            pass
        return False

    def getAttributeTypeInfo(self, catName, atName):
        try:
            return self.__tD[catName][atName]
        except Exception as e:
            if self.__verbose:
                logger.warning("Missing instance type info for category %r attribute %r %s" % (catName, atName, str(e)))
        return {}

    def getCategoryTypeInfo(self, catName):
        try:
            return self.__tD[catName]
        except Exception as e:
            if self.__verbose:
                logger.warning("Missing instance type info for category %r  %s" % (catName, str(e)))
        return {}

    def getMinWidth(self, catName, atName):
        try:
            return self.__tD[catName][atName]['minWidth']
        except Exception as e:
            if self.__verbose:
                logger.warning("Missing instance type info for category %r attribute %r %s" % (catName, atName, str(e)))
        return 0

    def getMaxWidth(self, catName, atName):
        try:
            return self.__tD[catName][atName]['maxWidth']
        except Exception as e:
            if self.__verbose:
                logger.warning("Missing instance type info for category %r attribute %r %s" % (catName, atName, str(e)))
        return 0

    def getMinPrecision(self, catName, atName):
        try:
            return self.__tD[catName][atName]['minPrec']
        except Exception as e:
            if self.__verbose:
                logger.warning("Missing instance type info for category %r attribute %r %s" % (catName, atName, str(e)))
        return 0

    def getMaxPrecision(self, catName, atName):
        try:
            return self.__tD[catName][atName]['maxPrec']
        except Exception as e:
            if self.__verbose:
                logger.warning("Missing instance type info for category %r attribute %r %s" % (catName, atName, str(e)))
        return 0

    def getCount(self, catName, atName):
        try:
            return self.__tD[catName][atName]['count']
        except Exception as e:
            if self.__verbose:
                logger.warning("Missing instance type info for category %r attribute %r %s" % (catName, atName, str(e)))
        return 0



##
# File:  Connection.py
# Date:  25-Mar-2018 J. Westbrook
#
# Update:
#   31-Mar-2018 jdw add context methods
#   23-Oct-2018 jdw add section name config access methods and make this a constructor argument
##
"""
Derived class for managing database credentials from a generic configuration file object.

"""
__docformat__ = "restructuredtext en"
__author__ = "John Westbrook"
__email__ = "jwest@rcsb.rutgers.edu"
__license__ = "Apache 2.0"


import logging
import platform

import MySQLdb

from rcsb.db.mysql.ConnectionBase import ConnectionBase

logger = logging.getLogger(__name__)
#
#
if platform.system() == "Linux":
    try:
        import sqlalchemy.pool as pool
        MySQLdb = pool.manage(MySQLdb, pool_size=12, max_overflow=12, timeout=30, echo=False, use_threadlocal=False)
    except Exception as e:
        logger.exception("Creating MYSQL connection pool failing with %s" % str(e))


class Connection(ConnectionBase):

    def __init__(self, cfgOb=None, infoD=None, resourceName=None, sectionName='site_server_info', verbose=False):
        super(Connection, self).__init__(verbose=verbose)
        #
        self.__cfgOb = cfgOb
        #
        if infoD:
            self.setPreferences(infoD)
        #
        if resourceName:
            self.assignResource(resourceName, sectionName)
        #

    def assignResource(self, resourceName=None, sectionName=None):
        """
        """
        #
        defaultPort = 3306
        dbServer = 'mysql'
        self._assignResource(resourceName)
        infoD = {}
        # if not self.__cfgOb:
        #    return infoD
        #
        if not resourceName or not sectionName:
            logger.exception("Missing resource specifiers resourceName %r sectionName %r" % (resourceName, sectionName))
        if (resourceName == "PRD"):
            infoD["DB_NAME"] = self.__cfgOb.get("SITE_REFDATA_PRD_DB_NAME", sectionName=sectionName)
            infoD["DB_HOST"] = self.__cfgOb.get("SITE_REFDATA_DB_HOST_NAME", sectionName=sectionName)
            infoD["DB_SOCKET"] = self.__cfgOb.get("SITE_REFDATA_DB_SOCKET", default=None, sectionName=sectionName)
            infoD["DB_PORT"] = self.__cfgOb.get("SITE_REFDATA_DB_PORT_NUMBER", default=defaultPort, sectionName=sectionName)

            infoD["DB_USER"] = self.__cfgOb.get("SITE_REFDATA_DB_USER_NAME", sectionName=sectionName)
            infoD["DB_PW"] = self.__cfgOb.get("SITE_REFDATA_DB_PASSWORD", sectionName=sectionName)

        elif (resourceName == "CC"):
            infoD["DB_NAME"] = self.__cfgOb.get("SITE_REFDATA_CC_DB_NAME", sectionName=sectionName)
            infoD["DB_HOST"] = self.__cfgOb.get("SITE_REFDATA_DB_HOST_NAME", sectionName=sectionName)
            infoD["DB_SOCKET"] = self.__cfgOb.get("SITE_REFDATA_DB_SOCKET", default=None, sectionName=sectionName)
            infoD["DB_PORT"] = self.__cfgOb.get("SITE_REFDATA_DB_PORT_NUMBER", default=defaultPort, sectionName=sectionName)

            infoD["DB_USER"] = self.__cfgOb.get("SITE_REFDATA_DB_USER_NAME", sectionName=sectionName)
            infoD["DB_PW"] = self.__cfgOb.get("SITE_REFDATA_DB_PASSWORD", sectionName=sectionName)

        elif (resourceName == "RCSB_INSTANCE"):
            infoD["DB_NAME"] = self.__cfgOb.get("SITE_INSTANCE_DB_NAME", sectionName=sectionName)
            infoD["DB_HOST"] = self.__cfgOb.get("SITE_INSTANCE_DB_HOST_NAME", sectionName=sectionName)
            infoD["DB_SOCKET"] = self.__cfgOb.get("SITE_INSTANCE_DB_SOCKET", default=None, sectionName=sectionName)
            infoD["DB_PORT"] = self.__cfgOb.get("SITE_INSTANCE_DB_PORT_NUMBER", default=defaultPort, sectionName=sectionName)

            self.__dbUser = self.__cfgOb.get("SITE_INSTANCE_DB_USER_NAME", sectionName=sectionName)
            infoD["DB_PW"] = self.__cfgOb.get("SITE_INSTANCE_DB_PASSWORD", sectionName=sectionName)

        elif (resourceName == "DA_INTERNAL"):
            infoD["DB_NAME"] = self.__cfgOb.get("SITE_DA_INTERNAL_DB_NAME", sectionName=sectionName)
            infoD["DB_HOST"] = self.__cfgOb.get("SITE_DA_INTERNAL_DB_HOST_NAME", sectionName=sectionName)
            infoD["DB_PORT"] = self.__cfgOb.get("SITE_DA_INTERNAL_DB_PORT_NUMBER", default=defaultPort, sectionName=sectionName)
            infoD["DB_SOCKET"] = self.__cfgOb.get("SITE_DA_INTERNAL_DB_SOCKET", default=None, sectionName=sectionName)

            infoD["DB_USER"] = self.__cfgOb.get("SITE_DA_INTERNAL_DB_USER_NAME", sectionName=sectionName)
            infoD["DB_PW"] = self.__cfgOb.get("SITE_DA_INTERNAL_DB_PASSWORD", sectionName=sectionName)

        elif (resourceName == "DA_INTERNAL_COMBINE"):
            infoD["DB_NAME"] = self.__cfgOb.get("SITE_DA_INTERNAL_COMBINE_DB_NAME", sectionName=sectionName)
            infoD["DB_HOST"] = self.__cfgOb.get("SITE_DA_INTERNAL_COMBINE_DB_HOST_NAME", sectionName=sectionName)
            infoD["DB_PORT"] = self.__cfgOb.get("SITE_DA_INTERNAL_COMBINE_DB_PORT_NUMBER", default=defaultPort, sectionName=sectionName)
            infoD["DB_SOCKET"] = self.__cfgOb.get("SITE_DA_INTERNAL_COMBINE_DB_SOCKET", default=None, sectionName=sectionName)

            infoD["DB_USER"] = self.__cfgOb.get("SITE_DA_INTERNAL_COMBINE_DB_USER_NAME", sectionName=sectionName)
            infoD["DB_PW"] = self.__cfgOb.get("SITE_DA_INTERNAL_COMBINE_DB_PASSWORD", sectionName=sectionName)
        elif (resourceName == "DISTRO"):
            infoD["DB_NAME"] = self.__cfgOb.get("SITE_DISTRO_DB_NAME", sectionName=sectionName)
            infoD["DB_HOST"] = self.__cfgOb.get("SITE_DISTRO_DB_HOST_NAME", sectionName=sectionName)
            infoD["DB_PORT"] = self.__cfgOb.get("SITE_DISTRO_DB_PORT_NUMBER", default=defaultPort, sectionName=sectionName)
            infoD["DB_SOCKET"] = self.__cfgOb.get("SITE_DISTRO_DB_SOCKET", default=None, sectionName=sectionName)

            infoD["DB_USER"] = self.__cfgOb.get("SITE_DISTRO_DB_USER_NAME", sectionName=sectionName)
            infoD["DB_PW"] = self.__cfgOb.get("SITE_DISTRO_DB_PASSWORD", sectionName=sectionName)

        elif (resourceName == "STATUS"):
            infoD["DB_NAME"] = self.__cfgOb.get("SITE_DB_DATABASE_NAME", sectionName=sectionName)
            infoD["DB_HOST"] = self.__cfgOb.get("SITE_DB_HOST_NAME", sectionName=sectionName)
            infoD["DB_PORT"] = self.__cfgOb.get("SITE_DB_PORT_NUMBER", default=defaultPort, sectionName=sectionName)
            infoD["DB_SOCKET"] = self.__cfgOb.get("SITE_DB_SOCKET", default=None, sectionName=sectionName)

            infoD["DB_USER"] = self.__cfgOb.get("SITE_DB_USER_NAME", sectionName=sectionName)
            infoD["DB_PW"] = self.__cfgOb.get("SITE_DB_PASSWORD", sectionName=sectionName)
        elif (resourceName == "MYSQL_DB"):

            infoD["DB_NAME"] = self.__cfgOb.get("MYSQL_DB_DATABASE_NAME", sectionName=sectionName)
            infoD["DB_HOST"] = self.__cfgOb.get("MYSQL_DB_HOST_NAME", sectionName=sectionName)
            infoD["DB_PORT"] = self.__cfgOb.get("MYSQL_DB_PORT_NUMBER", default=defaultPort, sectionName=sectionName)
            infoD["DB_SOCKET"] = self.__cfgOb.get("MYSQL_DB_SOCKET", default=None, sectionName=sectionName)

            infoD["DB_USER"] = self.__cfgOb.get("MYSQL_DB_USER_NAME", sectionName=sectionName)
            infoD["DB_PW"] = self.__cfgOb.get("MYSQL_DB_PASSWORD", sectionName=sectionName)
        else:
            pass

        infoD['DB_PORT'] = int(str(infoD['DB_PORT']))
        infoD['DB_SERVER'] = dbServer

        self.setPreferences(infoD)
        #
        return infoD

        #
    def __enter__(self):
        self.openConnection()
        return self.getClientConnection()

    def __exit__(self, *args):
        return self.closeConnection()



##
# File:    DictInfoHelperBase.py
# Author:  J. Westbrook
# Date:    18-May-2018
# Version: 0.001 Initial version
#
# Updates:
#    6-Jun-2018 jdw review the used
##
"""
Supplements dictionary information as required for schema production.

Single source of additional dictionary semantic content.

"""
__docformat__ = "restructuredtext en"
__author__ = "John Westbrook"
__email__ = "jwest@rcsb.rutgers.edu"
__license__ = "Apache 2.0"

import logging

logger = logging.getLogger(__name__)


class DictInfoHelperBase(object):
    """ Supplements standard dictionary content as required for schema production.

    """

    def __init__(self, **kwargs):
        self._raiseExceptions = kwargs.get('raiseExceptions', False)

    def getCardinalityKeyItem(self):
        """ Identify the parent item for the dictionary type that can be used to
            identify child categories with unity cardinality.   That is, logically containing
            a single data row in any instance.

        """
        if self._raiseExceptions:
            raise NotImplementedError()
        return ('', '')

    def getTypeCodes(self, kind):
        """
        """
        if self._raiseExceptions:
            raise NotImplementedError()
        return []

    def getQueryStrings(self, kind):
        if self._raiseExceptions:
            raise NotImplementedError()
        return []



##
# File: SequenceClustersEtlWorker.py
# Date: 2-Jul-2018  jdw
#
# ETL utilities for sequence cluster data
#
# Updates:
#  15-Jul-2018 jdw split out to separate module and add status tracking
#  28-Oct-2018 jdw adjustments for new configuration organization
#
##
__docformat__ = "restructuredtext en"
__author__ = "John Westbrook"
__email__ = "jwest@rcsb.rutgers.edu"
__license__ = "Apache 2.0"

import logging

from rcsb.db.mongo.DocumentLoader import DocumentLoader
from rcsb.db.processors.ClusterDataPrep import ClusterDataPrep
from rcsb.db.processors.DataExchangeStatus import DataExchangeStatus
from rcsb.db.utils.ProvenanceUtil import ProvenanceUtil

logger = logging.getLogger(__name__)


class SequenceClustersEtlWorker(object):
    """ Prepare and load sequence cluster data.

        Note: relevant configuration options -

        [entity_sequence_clusters]
        DATABASE_NAME=sequence_clusters
        DATABASE_VERSION_STRING=v5
        COLLECTION_ENTITY_MEMBERS=entity_members
        COLLECTION_ENTITY_MEMBERS_INDEX=data_set_id,entry_id,entity_id
        COLLECTION_CLUSTER_MEMBERS=cluster_members
        COLLECTION_CLUSTER_MEMBERS_INDEX=data_set_id,identity,cluster_id
        COLLECTION_VERSION_STRING=v0_1
        ENTITY_SCHEMA_NAME=rcsb_entity_sequence_cluster_entity_list
        CLUSTER_SCHEMA_NAME=rcsb_entity_sequence_cluster_identifer_list
        SEQUENCE_IDENTITY_LEVELS=100,95,90,70,50,30
        COLLECTION_CLUSTER_PROVENANCE=cluster_provenance
        PROVENANCE_KEY_NAME=rcsb_entity_sequence_cluster_prov
        PROVENANCE_INFO_LOCATOR=provenance/rcsb_extend_provenance_info.json


        """

    def __init__(self, cfgOb, workPath=None, numProc=2, chunkSize=10, readBackCheck=False, documentLimit=None, verbose=False):
        self.__cfgOb = cfgOb
        self.__workPath = workPath
        self.__readBackCheck = readBackCheck
        self.__numProc = numProc
        self.__chunkSize = chunkSize
        self.__documentLimit = documentLimit
        self.__resourceName = "MONGO_DB"
        self.__verbose = verbose
        #

        #
        self.__sectionCluster = 'entity_sequence_clusters'
        self.__clusterDataPath = self.__cfgOb.getPath('RCSB_SEQUENCE_CLUSTER_DATA_PATH', sectionName=self.__sectionCluster)
        self.__databaseName = self.__cfgOb.get('DATABASE_NAME', sectionName=self.__sectionCluster, default='sequence_clusters')
        self.__databaseVersion = self.__cfgOb.get('DATABASE_VERSION_STRING', sectionName=self.__sectionCluster, default='v5')
        #
        self.__entityMemberCollection = self.__cfgOb.get('COLLECTION_ENTITY_MEMBERS', sectionName=self.__sectionCluster, default='entity_members')
        self.__clusterMembersCollection = self.__cfgOb.get('COLLECTION_CLUSTER_MEMBERS', sectionName=self.__sectionCluster, default='cluster_members')
        self.__clusterProvenanceCollection = self.__cfgOb.get('COLLECTION_CLUSTER_PROVENANCE', sectionName=self.__sectionCluster, default='cluster_provenance')
        self.__collectionVersion = self.__cfgOb.get('COLLECTION_VERSION_STRING', sectionName=self.__sectionCluster, default='v0_1')
        #
        self.__entitySchemaName = self.__cfgOb.get('ENTITY_SCHEMA_NAME', sectionName=self.__sectionCluster, default='rcsb_entity_sequence_cluster_entity_list')
        self.__clusterSchemaName = self.__cfgOb.get('CLUSTER_SCHEMA_NAME', sectionName=self.__sectionCluster, default='rcsb_entity_sequence_cluster_identifer_list')
        #
        tS = self.__cfgOb.get('COLLECTION_ENTITY_MEMBERS_INDEX', sectionName=self.__sectionCluster, default=None)
        self.__entityMemberCollectionIndexL = tS.split(',') if tS else None
        tS = self.__cfgOb.get('COLLECTION_CLUSTER_MEMBERS_INDEX', sectionName=self.__sectionCluster, default=None)
        self.__clusterMembersCollectionIndexL = tS.split(',') if tS else None
        # sample data set
        #
        tS = self.__cfgOb.get('SEQUENCE_IDENTITY_LEVELS', sectionName=self.__sectionCluster, default=None)
        self.__identityLevels = tS.split(',') if tS else ['100', '95', '90', '70', '50', '30']
        #
        self.__statusList = []
        #

    def __updateStatus(self, updateId, databaseName, collectionName, status, startTimestamp):
        try:
            sFlag = 'Y' if status else 'N'
            desp = DataExchangeStatus()
            desp.setStartTime(tS=startTimestamp)
            desp.setObject(databaseName, collectionName)
            desp.setStatus(updateId=updateId, successFlag=sFlag)
            desp.setEndTime()
            self.__statusList.append(desp.getStatus())
            return True
        except Exception as e:
            logger.exception("Failing with %s" % str(e))
        return False

    def __extract(self, dataSetId, dataLocator, levels):
        """ Extract sequence cluster data set  (mmseq2 or blastclust organization)
        """
        try:
            cdp = ClusterDataPrep(workPath=self.__workPath, entitySchemaName=self.__entitySchemaName, clusterSchemaName=self.__clusterSchemaName)
            cifD, docBySequenceD, docByClusterD = cdp.extract(dataSetId, clusterSetLocator=dataLocator, levels=levels, clusterType='entity')
            return docBySequenceD, docByClusterD
        except Exception as e:
            logger.exception("Failing with %s" % str(e))

        return {}, {}

    def __fetchProvenance(self):
        """ Fetching a provenance dictionary content.
        """
        try:
            provKeyName = self.__cfgOb.get('PROVENANCE_KEY_NAME', sectionName=self.__sectionCluster, default='rcsb_entity_sequence_cluster_prov')
            provU = ProvenanceUtil(cfgOb=self.__cfgOb, workPath=self.__workPath)
            pD = provU.fetch(cfgSectionName=self.__sectionCluster)
            return pD[provKeyName] if provKeyName in pD else {}
        except Exception as e:
            logger.exception("Failing with %s" % str(e))
        return {}

    def etl(self, dataSetId, dataLocator=None, loadType='full'):
        """ Prepare and load sequence cluster data by entity and by cluster identifer.
        """
        try:
            self.__statusList = []
            desp = DataExchangeStatus()
            statusStartTimestamp = desp.setStartTime()
            #
            docBySequenceD, docByClusterD = self.__extract(dataSetId=dataSetId, dataLocator=dataLocator, levels=self.__identityLevels)
            #
            dl = DocumentLoader(self.__cfgOb, self.__resourceName, numProc=self.__numProc, chunkSize=self.__chunkSize,
                                documentLimit=self.__documentLimit, verbose=self.__verbose, readBackCheck=self.__readBackCheck)
            #
            databaseName = self.__databaseName + '_' + self.__databaseVersion
            #
            collectionName = self.__entityMemberCollection + '_' + self.__collectionVersion
            dList = docBySequenceD[self.__entitySchemaName]
            ok1 = dl.load(databaseName, collectionName, loadType=loadType, documentList=dList, indexAttributeList=self.__entityMemberCollectionIndexL, keyNames=None)
            self.__updateStatus(dataSetId, databaseName, collectionName, ok1, statusStartTimestamp)
            #
            collectionName = self.__clusterMembersCollection + '_' + self.__collectionVersion
            dList = docByClusterD[self.__clusterSchemaName]
            ok2 = dl.load(databaseName, collectionName, loadType=loadType, documentList=dList, indexAttributeList=self.__clusterMembersCollectionIndexL, keyNames=None)
            self.__updateStatus(dataSetId, databaseName, collectionName, ok2, statusStartTimestamp)
            #
            pD = self.__fetchProvenance()
            collectionName = self.__clusterProvenanceCollection + '_' + self.__collectionVersion
            ok3 = dl.load(databaseName, collectionName, loadType=loadType, documentList=[pD], indexAttributeList=None, keyNames=None)
            self.__updateStatus(dataSetId, databaseName, collectionName, ok3, statusStartTimestamp)
            #
            return ok1 and ok2 and ok3
        except Exception as e:
            logger.exception("Failing with %s" % str(e))
        return False

    def getLoadStatus(self):
        return self.__statusList



##
# File:  ConnectionBase.py
# Date:  25-Mar-2018 J. Westbrook
#
# Update:
##
"""
Base class for managing database connection for MySQL.  Application credentials are
handled by the derived class.

"""
__docformat__ = "restructuredtext en"
__author__ = "John Westbrook"
__email__ = "jwest@rcsb.rutgers.edu"
__license__ = "Apache 2.0"


import copy
import logging
import platform

import MySQLdb

logger = logging.getLogger(__name__)
#
#
if platform.system() == "Linux":
    try:
        import sqlalchemy.pool as pool
        MySQLdb = pool.manage(MySQLdb, pool_size=12, max_overflow=12, timeout=30, echo=False, use_threadlocal=False)
    except Exception as e:
        logger.exception("Creating MYSQL connection pool failing with %s" % str(e))


class ConnectionBase(object):

    def __init__(self, siteId=None, verbose=False):
        self.__verbose = verbose
        #
        self.__siteId = siteId

        self.__db = None
        self._dbCon = None

        self.__infoD = {}
        self.__databaseName = None
        self.__dbHost = None
        self.__dbUser = None
        self.__dbPw = None
        self.__dbSocket = None
        self.__dbPort = None
        self.__dbAdminDb = None
        self.__dbPort = None
        self.__defaultPort = 3306
        self.__dbServer = 'mysql'
        self.__resourceName = None

    def assignResource(self, resourceName=None):
        # implement in the derived class
        self._assignResource(resourceName)

    def _assignResource(self, resourceName):
        self.__resourceName = resourceName

    def getPreferences(self):
        return self.__infoD

    def setPreferences(self, infoD):
        try:
            self.__infoD = copy.deepcopy(infoD)
            self.__databaseName = self.__infoD.get("DB_NAME", None)
            self.__dbHost = self.__infoD.get("DB_HOST", 'localhost')
            self.__dbUser = self.__infoD.get("DB_USER", None)
            self.__dbPw = self.__infoD.get("DB_PW", None)
            self.__dbSocket = self.__infoD.get("DB_SOCKET", None)
            self.__dbServer = self.__infoD.get("DB_SERVER", "mysql")
            #
            port = self.__infoD.get("DB_PORT", self.__defaultPort)
            if port and len(str(port)) > 0:
                self.__dbPort = int(str(port))
        except Exception as e:
            logger.exception("Failing with %s" % str(e))

    def openConnection(self):
        """ Create a database connection and return a connection object.

            Returns None on failure
        """
        #
        if self._dbCon is not None:
            # Close an open connection -
            logger.info("+MyDbConnect.connect() WARNING Closing an existing connection.")
            self.closeConnection()

        try:
            if self.__dbSocket is None:
                dbcon = MySQLdb.connect(db="%s" % self.__databaseName,
                                        user="%s" % self.__dbUser,
                                        passwd="%s" % self.__dbPw,
                                        host="%s" % self.__dbHost,
                                        port=self.__dbPort,
                                        local_infile=1)
            else:
                dbcon = MySQLdb.connect(db="%s" % self.__databaseName,
                                        user="%s" % self.__dbUser,
                                        passwd="%s" % self.__dbPw,
                                        host="%s" % self.__dbHost,
                                        port=self.__dbPort,
                                        unix_socket="%s" % self.__dbSocket,
                                        local_infile=1)

            self._dbCon = dbcon
            return True
        except Exception as e:
            logger.exception("Connection error to resource %s with %s" % (self.__resourceName, str(e)))
            self._dbCon = None

        return False

    def getClientConnection(self):
        return self._dbCon

    def closeConnection(self):
        """ Close db session
        """
        if self._dbCon is not None:
            self._dbCon.close()
            self._dbCon = None
            return True
        else:
            return False

    def getCursor(self):
        try:
            return self._dbCon.cursor()
        except Exception as e:
            logger.exception("Failing with %s" % str(e))

        return None



##
# File:    CrateDbUtil.py
# Author:  J. Westbrook
# Date:    21-Dec-2017
# Version: 0.001 Initial version
#
# Updates:
#
##
"""
Utility classes to create connections and process SQL commands with CrateDb.

"""
from __future__ import generators

import logging

from crate.client.exceptions import (DatabaseError, OperationalError,
                                     ProgrammingError, Warning)

__docformat__ = "restructuredtext en"
__author__ = "John Westbrook"
__email__ = "jwest@rcsb.rutgers.edu"
__license__ = "Apache 2.0"

logger = logging.getLogger(__name__)


class CrateDbQuery(object):

    """ Parameterized SQL queries using Python DBI protocol...
    """

    def __init__(self, dbcon, verbose=True):
        self.__dbcon = dbcon
        self.__verbose = verbose
        self.__ops = ['EQ', 'GE', 'GT', 'LT', 'LE', 'LIKE', 'NOT LIKE']
        self.__opDict = {'EQ': '=',
                         'GE': '>=',
                         'GT': '>',
                         'LT': '<',
                         'LE': '<=',
                         'LIKE': 'LIKE',
                         'NOT LIKE': 'NOT LIKE'
                         }
        self.__logOps = ['AND', 'OR', 'NOT']
        self.__grpOps = ['BEGIN', 'END']
        self.__warningAction = 'default'

    def sqlTemplateCommandMany(self, sqlTemplate, valueLists=None):
        """  Execute a batch sql commands followed by a single commit. Commands are
             are describe in a template with an associated list of values.

            cursor.executemany("INSERT INTO locations (name, date, kind, position) VALUES (?, ?, ?, ?)",
                ...                [('Cloverleaf', '2007-03-11', 'Quasar', 7),
                ...                 ('Old Faithful', '2007-03-11', 'Quasar', 7)])
                [{u'rowcount': 1}, {u'rowcount': 1}]

        """
        lenR = 0
        ret = []
        iFail = 0
        try:
            lenT = len(valueLists)
            curs = self.__dbcon.cursor()
            ret = curs.executemany(sqlTemplate, valueLists)
            lenR = len(ret)
            logger.debug("Return len %d" % len(ret))
            for i in range(len(ret)):
                if ret[i]['rowcount'] != 1:
                    iFail += 1
                    logger.info("Insert fails on row %d of %d with values: %r" % (i, lenT, valueLists[i]))
            curs.close()
            return lenR - iFail
        except DatabaseError as e:
            logger.info("sqlTemplate %s" % sqlTemplate)
            logger.info("return list %r" % ret)
            logger.error("error is:\n%s" % str(e))
            curs.close()
        except Warning as e:
            logger.warning("warning is:\n%s" % str(e))
            curs.close()
        except Exception as e:
            logger.exception("Exception is:\n%s" % str(e))
            curs.close()
        #
        lenR = len(ret)
        return lenR - iFail

    def sqlTemplateCommand(self, sqlTemplate=None, valueList=[]):
        """  Execute sql template command with associated value list.

             Insert one row -

             Errors and warnings that generate exceptions are caught by this method.
        """
        try:
            curs = self.__dbcon.cursor()
            curs.execute(sqlTemplate, valueList)
            curs.close()
            return True
        except DatabaseError as e:
            logger.info(" error is:\n%s\n" % str(e))
            curs.close()
        except Warning as e:
            logger.info(" warning is:\n%s\n" % str(e))
            curs.close()
        except Exception as e:
            logger.info(" exception is:\n%s\n" % str(e))
            curs.close()
        return False

    def sqlTemplateCommandList(self, sqlTemplateValueList=None):
        """  Execute sql template command with associated value list.

             Input -

             sqlTemplateValueList [(sqlTemplate,vList), (sqlTemplate, vlist), ... ]

             Insert on row -

             Errors and warnings that generate exceptions are caught by this method.
        """
        vL = []
        iFail = 0
        try:
            curs = self.__dbcon.cursor()
            #
            lenT = len(sqlTemplateValueList)
            for ii in range(lenT):
                t, vL = sqlTemplateValueList[ii]
                try:
                    curs.execute(t, vL)
                except Exception as e:
                    iFail += 1
                    logger.info(" Error is: %s" % str(e))
                    # logger.info(" Template for record %d of %d : %s" % (ii, lenT, t))
                    logger.info(" Record %d of %d value list: %s" % (ii, lenT, vL))
            #
            curs.close()
            logger.debug(" Inserted %d of %d values" % (ii - iFail, lenT))
            return ii - iFail + 1
        except DatabaseError as e:
            logger.exception(" error is: %s" % str(e))
            logger.info(" Record %d of %d value list: %s" % (ii, lenT, vL))
            curs.close()
        except Warning as e:
            logger.info(" Warning is: %s" % str(e))
            logger.info(" Record %d of %d value list: %s" % (ii, lenT, vL))
            curs.close()
        except Exception as e:
            logger.info(" Exception is: %s" % str(e))
            logger.info(" Record %d of %d value list: %s" % (ii, lenT, vL))
            curs.close()
        return ii - iFail + 1

    def sqlCommandList(self, sqlCommandList):
        """  Execute the input list of SQL commands catching exceptions from the server.

        The treatment of warning is controlled by a prior setting of self.setWarnings("error"|"ignore"|"default")
        """

        try:
            sqlCommand = ''
            curs = self.__dbcon.cursor()
            for sqlCommand in sqlCommandList:
                curs.execute(sqlCommand)
            #
            curs.close()
            return True
        except DatabaseError as e:
            logger.info(" SQL command failed for:\n%s" % sqlCommand)
            logger.info(" database error is message is:\n%s" % str(e))
            curs.close()
        except Warning as e:
            logger.info(" warning message is:\n%s" % str(e))
            logger.info(" generated warnings for command:\n%s" % sqlCommand)
            curs.close()
        except Exception as e:
            logger.info(" exception message is:\n%s\n" % str(e))
            logger.exception(" SQL command failed for:\n%s\n" % sqlCommand)
            curs.close()

        return False

    def sqlCommand(self, queryString):
        """   Execute SQL command catching exceptions returning no data from the server.
        """
        try:
            curs = self.__dbcon.cursor()
            curs.execute(queryString)
            curs.close()
            return True
        except OperationalError as e:
            logger.info(" SQL command failed for:\n%s" % queryString)
            logger.info(" warning is message is:\n%s" % str(e))
            curs.close()
        except DatabaseError as e:
            logger.info(" SQL command failed for:\n%s\n" % queryString)
            logger.info(" MySQL warning is message is:\n%s\n" % str(e))
            curs.close()
        except Exception as e:
            logger.exception(" SQL command failed for:\n%s\n with %s" % (queryString, str(e)))
            curs.close()
        return []

    def __fetchIter(self, cursor, rowSize=1000):
        """ Chunked iterator to manage results fetches to mysql server
        """
        while True:
            results = cursor.fetchmany(rowSize)
            if not results:
                break
            for result in results:
                yield result

    def selectRows(self, queryString):
        """ Execute SQL command and return list of lists for the result set.
        """
        rowList = []
        try:
            curs = self.__dbcon.cursor()
            curs.execute(queryString)
            while True:
                result = curs.fetchone()
                if (result is not None):
                    rowList.append(result)
                else:
                    break
            curs.close()
            return rowList
        except ProgrammingError as e:
            logger.info(" MySQL warning is message is:\n%s\n" % str(e))
            curs.close()
        except OperationalError as e:
            logger.info(" MySQL warning is message is:\n%s\n" % str(e))
            logger.info(" SQL command failed for:\n%s\n" % queryString)
            curs.close()
        except DatabaseError as e:
            logger.info(" MySQL warning is message is:\n%s\n" % str(e))
            logger.info(" SQL command failed for:\n%s\n" % queryString)
            curs.close()
        except Exception as e:
            logger.exception(" SQL command failed for:\n%s\n with %s" % (queryString, str(e)))
            curs.close()

        return []

    def simpleQuery(self, selectList=[], fromList=[], condition='',
                    orderList=[], returnObj=[]):
        """
        """
        #
        colsCsv = ",".join(["%s" % k for k in selectList])
        tablesCsv = ",".join(["%s" % k for k in fromList])

        order = ""
        if (len(orderList) > 0):
            (a, t) = orderList[0]
            order = " ORDER BY CAST(%s AS %s) " % (a, t)
            for (a, t) in orderList[1:]:
                order += ", CAST(%s AS %s) " % (a, t)

        #
        query = "SELECT " + colsCsv + " FROM " + tablesCsv + condition + order
        if (self.__verbose):
            logger.info("Query: %s\n" % query)
        curs = self.__dbcon.cursor()
        curs.execute(query)
        while True:
            result = curs.fetchone()
            if (result is not None):
                returnObj.append(result)
            else:
                break
        curs.close()
        return returnObj

    def testSelectQuery(self, count):
        tSQL = "select %d" % count
        #
        try:
            rowL = self.selectRows(queryString=tSQL)
            tup = rowL[0]
            return int(str(tup[0])) == count
        except Exception:
            return False



##
# File: RepoHoldingsDataPrep.py
# Author:  J. Westbrook
# Date:  12-Jul-2018
#
# Note:  This module uses well sandbox and repository defined file names
#        within the configuration defined RCSB exchange sandbox path.
#
# Update:
# 16-Jul-2018 jdw adjust naming to current sandbox conventions.
# 30-Oct-2018 jdw naming to conform to dictioanry and schema conventions.
#  8-Oct-2018 jdw adjustments for greater schema compliance and uniformity
#                 provide alternative date assignment for JSON and mongo validation
# 10-Oct-2018 jdw Add getHoldingsTransferred() for homology models transfered to Model Archive
# 28-Oct-2018 jdw update with semi-colon separated author lists in status and theoretical model files
# 25-Nov-2018 jdw add sequence/pdb_seq_prerelease.fasta and  _rcsb_repository_holdings_prerelease.seq_one_letter_code
# 27-Nov-2018 jdw update rcsb_repository_holdings_current for all entry content types
# 29-Nov-2018 jdw Add support for NMR restraint versions.
# 30-Nov-3018 jdw explicitly filter obsolete entries from current holdings
# 13-Dec-2018 jdw Adjust logic for reporting assembly format availibility
##

__docformat__ = "restructuredtext en"
__author__ = "John Westbrook"
__email__ = "jwest@rcsb.rutgers.edu"
__license__ = "Apache 2.0"


import logging
import os

import dateutil.parser

from rcsb.utils.io.MarshalUtil import MarshalUtil

logger = logging.getLogger(__name__)


class RepoHoldingsDataPrep(object):
    """
    Consolidate legacy data describing repository content updates and repository entry status.

    """

    def __init__(self, **kwargs):
        self.__workPath = kwargs.get('workPath', None)
        self.__sandboxPath = kwargs.get('sandboxPath', None)
        self.__filterType = kwargs.get('filterType', '')
        self.__assignDates = 'assign-dates' in self.__filterType
        #
        self.__mU = MarshalUtil(workPath=self.__workPath)
        #

    def getHoldingsTransferred(self, updateId, dirPath=None, **kwargs):
        """ Parse legacy lists defining the repository contents transferred to alternative repositories

        Args:
            updateId (str): update identifier (e.g. 2018_32)
            dirPath (str): directory path containing update list files
            **kwargs: unused

        Returns:
            list: List of dictionaries containing data for rcsb_repository_holdings_transferred
            list: List of dictionaries containing data for rcsb_repository_holdings_insilico_models

        Example input data:

        ma-czyyf : 262D - TITLE A THREE-DIMENSIONAL MODEL OF THE REV BINDING ELEMENT OF HIV- TITLE 2 1 DERIVED FROM ANALYSES OF IN VITRO SELECTED VARIANTS
        ma-cfqla : 163D - TITLE A THREE-DIMENSIONAL MODEL OF THE REV BINDING ELEMENT OF HIV- TITLE 2 1 DERIVED FROM ANALYSES OF IN VITRO SELECTED VARIANTS

        and -

        1DX2    REL 1999-12-16  2000-12-15  Tumour Targetting Human ...  Beiboer, S.H.W., Reurs, A., Roovers, R.C., Arends, J., Whitelegg, N.R.J., Rees, A.R., Hoogenboom, H.R.

        and -

        1APD    OBSLTE  1992-10-15      2APD
        1BU0    OBSLTE  1998-10-07      2BU0
        1CLJ    OBSLTE  1998-03-04      2CLJ
        1DU8    OBSLTE  2001-01-31      1GIE
        1I2J    OBSLTE  2001-01-06      1JA5

        """
        trsfL = []
        insL = []
        dirPath = dirPath if dirPath else self.__sandboxPath
        try:
            fp = os.path.join(dirPath, 'status', 'theoretical_model_obsolete.tsv')
            lineL = self.__mU.doImport(fp, 'list')
            #
            obsDateD = {}
            obsIdD = {}
            for line in lineL:
                fields = line.split('\t')
                if len(fields) < 3:
                    continue
                entryId = str(fields[0]).strip().upper()
                obsDateD[entryId] = dateutil.parser.parse(fields[2]) if self.__assignDates else fields[2]
                if len(fields) > 3 and len(fields[3]) > 3:
                    obsIdD[entryId] = fields[3]
            logger.debug("Read %d obsolete insilico id codes" % len(obsDateD))
            # ---------  ---------  ---------  ---------  ---------  ---------  ---------
            fp = os.path.join(dirPath, 'status', 'model-archive-PDB-insilico-mapping.list')
            lineL = self.__mU.doImport(fp, 'list')
            #
            trD = {}
            for line in lineL:
                fields = line.split(':')
                if len(fields) < 2:
                    continue
                entryId = str(fields[1]).strip().upper()[:4]
                maId = str(fields[0]).strip()
                trD[entryId] = maId
            logger.debug("Read %d model archive id codes" % len(trD))
            #
            # ---------  ---------  ---------  ---------  ---------  ---------  ---------
            fp = os.path.join(dirPath, 'status', 'theoretical_model_v2.tsv')
            lineL = self.__mU.doImport(fp, 'list')
            #
            logger.debug("Read %d insilico id codes" % len(lineL))
            for line in lineL:
                fields = str(line).split('\t')
                if len(fields) < 6:
                    continue
                depDate = dateutil.parser.parse(fields[2]) if self.__assignDates else fields[2]
                relDate = None
                if len(fields[3]) >= 10 and not fields[3].startswith('0000'):
                    relDate = dateutil.parser.parse(fields[3]) if self.__assignDates else fields[3]

                statusCode = 'TRSF' if fields[1] == 'REL' else fields[1]

                entryId = str(fields[0]).upper()
                title = fields[4]
                #
                auditAuthors = [t.strip() for t in fields[5].split(';')]
                repId = None
                if entryId in trD:
                    repName = 'Model Archive'
                    repId = trD[entryId]

                #
                d = {'update_id': updateId,
                     'entry_id': entryId,
                     'status_code': statusCode,
                     'deposit_date': depDate,
                     'repository_content_types': ['coordinates'],
                     'title': title,
                     'audit_authors': auditAuthors}
                #
                if relDate:
                    d['release_date'] = relDate
                #
                if repId:
                    d['remote_accession_code'] = repId
                    d['remote_repository_name'] = repName
                if statusCode == 'TRSF':
                    trsfL.append(d)
                #
                #
                d = {'update_id': updateId,
                     'entry_id': entryId,
                     'status_code': statusCode,
                     'deposit_date': depDate,
                     'title': title,
                     'audit_authors': auditAuthors}
                #
                if relDate:
                    d['release_date'] = relDate
                #
                if entryId in obsDateD:
                    d['remove_date'] = relDate
                #
                if entryId in obsIdD:
                    d['id_codes_replaced_by'] = [obsIdD[entryId]]
                #
                insL.append(d)
            #
        except Exception as e:
            logger.exception("Failing with %s" % str(e))

        return trsfL, insL

    def getHoldingsUpdate(self, updateId, dirPath=None, **kwargs):
        """ Parse legacy lists defining the contents of the repository update

        Args:
            updateId (str): update identifier (e.g. 2018_32)
            dirPath (str): directory path containing update list files
            **kwargs: unused

        Returns:
            list: List of dictionaries containing rcsb_repository_holdings_update
        """
        retL = []
        dirPath = dirPath if dirPath else self.__sandboxPath
        try:
            updateTypeList = ['added', 'modified', 'obsolete']
            contentTypeList = ['entries', 'mr', 'cs', 'sf']
            contentNameD = {'entries': 'coordinates', 'mr': 'NMR restraints', 'cs': 'NMR chemical shifts', 'sf': 'structure factors'}
            #
            for updateType in updateTypeList:
                for contentType in contentTypeList:
                    fp = os.path.join(dirPath, 'update-lists', updateType + '-' + contentType)
                    entryIdL = self.__mU.doImport(fp, 'list')
                    #
                    uD = {}
                    for entryId in entryIdL:
                        entryId = entryId.strip().upper()
                        if entryId not in uD:
                            uD[entryId] = []
                        uD[entryId].append(contentNameD[contentType])
                for entryId in uD:
                    uType = 'removed' if updateType == 'obsolete' else updateType
                    retL.append({'update_id': updateId, 'entry_id': entryId, 'update_type': uType, 'repository_content_types': uD[entryId]})
            return retL
        except Exception as e:
            logger.exception("Failing with %s" % str(e))

        return retL

    def getHoldingsCurrent(self, updateId, dirPath=None, **kwargs):
        """ Parse legacy lists defining the current contents of the repository update

        Args:
            updateId (str): update identifier (e.g. 2018_32)
            dirPath (str): directory path containing update list files
            **kwargs: unused

        Returns:
            list: List of dictionaries containing data for rcsb_repository_holdings_current
        """
        rD = {}
        retL = []
        dirPath = dirPath if dirPath else self.__sandboxPath
        try:
            updateTypeList = ['all']
            contentTypeList = ['pdb', 'mr', 'cs', 'sf']
            contentNameD = {'pdb': 'coordinates', 'mr': 'NMR restraints', 'cs': 'NMR chemical shifts', 'sf': 'structure factors'}
            #
            tD = {}
            for updateType in updateTypeList:
                for contentType in contentTypeList:
                    fp = os.path.join(dirPath, 'update-lists', updateType + '-' + contentType + '-list')
                    entryIdL = self.__mU.doImport(fp, 'list')
                    #
                    for entryId in entryIdL:
                        entryId = entryId.strip().upper()
                        if entryId not in tD:
                            tD[entryId] = {}
                        tD[entryId][contentNameD[contentType]] = True
            #
            fp = os.path.join(dirPath, 'status', 'biounit_file_list.tsv')
            lines = self.__mU.doImport(fp, 'list')
            assemD = {}
            for line in lines:
                fields = line.split('\t')
                entryId = fields[0].strip().upper()
                assemId = fields[1].strip()
                if entryId not in assemD:
                    assemD[entryId] = []
                assemD[entryId].append(assemId)
            #
            #
            fp = os.path.join(dirPath, 'status', 'pdb_bundle_index_list.tsv')
            bundleIdList = self.__mU.doImport(fp, 'list')
            bundleD = {}
            for entryId in bundleIdList:
                bundleD[entryId.strip().upper()] = True
            #
            fp = os.path.join(dirPath, 'status', 'validation_report_list.tsv')
            vList = self.__mU.doImport(fp, 'list')
            valD = {}
            for entryId in vList:
                valD[entryId.strip().upper()] = True
            #
            #
            fp = os.path.join(dirPath, 'status', 'entries_without_polymers.tsv')
            pList = self.__mU.doImport(fp, 'list')
            pD = {}
            for entryId in pList:
                pD[entryId.strip().upper()] = False
            #
            #
            fp = os.path.join(dirPath, 'status', 'nmr_restraints_v2_list.tsv')
            nmrV2List = self.__mU.doImport(fp, 'list')
            nmrV2D = {}
            for entryId in nmrV2List:
                nmrV2D[entryId.strip().upper()] = False
            #
            fp = os.path.join(dirPath, 'status', 'edmaps.json')
            qD = self.__mU.doImport(fp, 'json')
            edD = {}
            for entryId in qD:
                edD[entryId.upper()] = qD[entryId]
            #
            fp = os.path.join(dirPath, 'status', 'obsolete_entry.json_2')
            oL = self.__mU.doImport(fp, 'json')
            obsD = {}
            for d in oL:
                obsD[d['entryId'].upper()] = True
            logger.info("Removed entry length %d" % len(obsD))
            #
            #
            # Revise content types bundles and assemblies
            #
            for entryId, dD in tD.items():
                if entryId in obsD:
                    continue
                rD[entryId] = []
                if 'coordinates' in dD and entryId in bundleD:
                    rD[entryId].append('entry PDB bundle')
                    rD[entryId].append('entry mmCIF')
                    rD[entryId].append('entry PDBML')
                else:
                    rD[entryId].append('entry PDB')
                    rD[entryId].append('entry mmCIF')
                    rD[entryId].append('entry PDBML')
                if entryId in assemD:
                    if entryId in bundleD:
                        rD[entryId].append('assembly mmCIF')
                    else:
                        rD[entryId].append('assembly PDB')
                #
                for cType in dD:
                    if cType not in ['coordinates', 'NMR restraints']:
                        rD[entryId].append(cType)
                    if cType == 'NMR restraints':
                        rD[entryId].append('NMR restraints V1')

                if entryId in nmrV2D:
                    rD[entryId].append('NMR restraints V2')
                #
                if entryId in valD:
                    rD[entryId].append('validation report')
                if entryId in edD:
                    rD[entryId].append('2fo-fc Map')
                    rD[entryId].append('fo-fc Map')
                    rD[entryId].append('Map Coefficients')
                if entryId not in pD:
                    rD[entryId].append('FASTA sequence')
            #
            for entryId in rD:
                if entryId in assemD:
                    retL.append({'update_id': updateId, 'entry_id': entryId, 'assembly_ids': assemD[entryId], 'repository_content_types': rD[entryId]})
                else:
                    retL.append({'update_id': updateId, 'entry_id': entryId, 'repository_content_types': rD[entryId]})
            return retL
        except Exception as e:
            logger.exception("Failing with %s" % str(e))

        return retL

    def getHoldingsUnreleased(self, updateId, dirPath=None, **kwargs):
        """ Parse the legacy exchange status file containing details for unreleased entries:

        Args:
            updateId (str): update identifier (e.g. 2018_32)
            dirPath (str): directory path containing update list files
            **kwargs: unused

        Returns:
            list: List of dictionaries containing data for rcsb_repository_holdings_unreleased

        """
        retL = []
        fields = []
        dirPath = dirPath if dirPath else self.__sandboxPath
        try:
            #
            fp = os.path.join(dirPath, 'status', 'status_v2.txt')
            lines = self.__mU.doImport(fp, 'list')
            for line in lines:
                fields = line.split('\t')
                if len(fields) < 15:
                    continue
                entryId = fields[1]
                d = {'update_id': updateId,
                     'entry_id': entryId,
                     'status_code': fields[2]
                     # 'sg_project_name': fields[14],
                     # 'sg_project_abbreviation_': fields[15]}
                     }
                if fields[11] and len(fields[11].strip()):
                    d['title'] = fields[11]
                if fields[10] and len(fields[10].strip()):
                    d['audit_authors'] = [t.strip() for t in fields[10].split(';')]
                    # d['audit_authors'] = fields[10]
                if fields[12] and len(fields[12].strip()):
                    d['author_prerelease_sequence'] = str(fields[12]).strip()
                dTupL = [('deposit_date', 3),
                         ('deposit_date_coordinates', 4),
                         ('deposit_date_structure_factors', 5),
                         ('hold_date_struct_fact', 6),
                         ('deposit_date_nmr_restraints', 7),
                         ('hold_date_nmr_restraints', 8),
                         ('release_date', 9),
                         ('hold_date_coordinates', 13)]
                for dTup in dTupL:
                    fN = dTup[1]
                    if fields[fN] and len(fields[fN]) >= 4:
                        d[dTup[0]] = dateutil.parser.parse(fields[fN]) if self.__assignDates else fields[fN]
                #
                retL.append({k: v for k, v in d.items() if v})
        except Exception as e:
            logger.error("Fields: %r" % fields)
            logger.exception("Failing with %s" % str(e))

        return retL

    def getHoldingsRemoved(self, updateId, dirPath=None, **kwargs):
        """ Parse the legacy exchange file containing details of removed entries:

            {
                "entryId": "125D",
                "obsoletedDate": "1998-04-15",
                "title": "SOLUTION STRUCTURE OF THE DNA-BINDING DOMAIN OF CD=2=-GAL4 FROM S. CEREVISIAE",
                "details": "",
                "depositionAuthors": [
                    "Baleja, J.D.",
                    "Wagner, G."
                ],
                "depositionDate": "1993-05-05",
                "releaseDate": "1994-01-31",
                "obsoletedBy": [
                    "1AW6"
                ]},

        Returns;
            (list) : list of dictionaries for rcsb_repository_holdings_removed
            (list) : list of dictionaries for rcsb_repository_holdings_removed_audit_authors
            (list) : list of dictionaries for rcsb_repository_holdings_superseded

        """
        # rcsb_repository_holdings_removed
        rL1 = []
        # rcsb_repository_holdings_removed_audit_authors
        rL2 = []
        # rcsb_repository_holdings_superseded
        rL3 = []
        #
        sD = {}
        dirPath = dirPath if dirPath else self.__sandboxPath
        try:
            fp = os.path.join(dirPath, 'status', 'obsolete_entry.json_2')
            dD = self.__mU.doImport(fp, 'json')
            for d in dD:
                rbL = d['obsoletedBy'] if 'obsoletedBy' in d else []
                d1 = {'update_id': updateId,
                      'entry_id': d['entryId'],
                      'title': d['title'],
                      'details': d['details'],
                      'audit_authors': d['depositionAuthors']}
                if rbL:
                    d1['id_codes_replaced_by'] = rbL

                dTupL = [('deposit_date', 'depositionDate'),
                         ('remove_date', 'obsoletedDate'),
                         ('release_date', 'releaseDate')]
                for dTup in dTupL:
                    fN = dTup[1]
                    if d[fN] and len(d[fN]) > 4:
                        d1[dTup[0]] = dateutil.parser.parse(d[fN]) if self.__assignDates else d[fN]

                rL1.append({k: v for k, v in d1.items() if v})
                #
                for ii, author in enumerate(d['depositionAuthors']):
                    d2 = {'update_id': updateId, 'entry_id': d['entryId'], 'ordinal_id': ii + 1, 'audit_author': author}
                    rL2.append(d2)
                if 'obsoletedBy' in d:
                    for pdbId in d['obsoletedBy']:
                        if pdbId not in sD:
                            sD[pdbId] = []
                        sD[pdbId].append(d['entryId'])
            for pdbId in sD:
                if len(sD[pdbId]) > 1:
                    rL3.append({'update_id': updateId, 'entry_id': pdbId, 'id_codes_superseded': sD[pdbId]})

            logger.debug("rl3 %r" % rL3)
            logger.debug("Computed data lengths  %d %d %d" % (len(rL1), len(rL2), len(rL3)))
        except Exception as e:
            logger.exception("Failing with %s" % str(e))

        return rL1, rL2, rL3

    def getHoldingsPrerelease(self, updateId, dirPath=None, **kwargs):
        """ Parse the legacy exchange status file containing prerelease sequence data.

        Args:
            updateId (str): update identifier (e.g. 2018_32)
            dirPath (str): directory path containing update list files
            **kwargs: unused

        Returns:
            list: List of dictionaries containing data for rcsb_repository_holdings_prerelease

        """
        retL = []
        fields = []
        dirPath = dirPath if dirPath else self.__sandboxPath
        try:
            # Get prerelease sequence data
            fp = os.path.join(dirPath, 'sequence', 'pdb_seq_prerelease.fasta')
            sD = self.__mU.doImport(fp, 'fasta', commentStyle="prerelease")
            seqD = {}
            for sid in sD:
                fields = sid.split('_')
                entryId = str(fields[0]).upper()
                if entryId not in seqD:
                    seqD[entryId] = []
                seqD[entryId].append(sD[sid]['sequence'])
            logger.debug("Loaded prerelease sequences for %d entries" % len(seqD))
            #
            for entryId, seqL in seqD.items():
                d = {'update_id': updateId, 'entry_id': entryId, 'seq_one_letter_code': seqL}
                logger.debug("Adding prerelease sequences for %s" % entryId)
                retL.append({k: v for k, v in d.items() if v})
        except Exception as e:
            logger.error("Fields: %r" % fields)
            logger.exception("Failing with %s" % str(e))

        return retL



##
# File: DataExchangeStatus.py
# Author:  J. Westbrook
# Date:  10-Jul-2018
#
# Update:
# 14-Jul-2018 jdw update docs and return datetime objects for timestamp strings.
# 14-Jul-2018 jdw return timestamps from set methods
##

__docformat__ = "restructuredtext en"
__author__ = "John Westbrook"
__email__ = "jwest@rcsb.rutgers.edu"
__license__ = "Apache 2.0"


import logging

from rcsb.db.utils.TimeUtil import TimeUtil

logger = logging.getLogger(__name__)


class DataExchangeStatus(object):
    """
    Create status records for data exchange operations.

    For example,

    loop_
     _rcsb_data_exchange_status.update_id
     _rcsb_data_exchange_status.database
     _rcsb_data_exchange_status.object
     _rcsb_data_exchange_status.update_status_flag
     _rcsb_data_exchange_status.update_begin_timestamp
     _rcsb_data_exchange_status.update_end_timestamp
    2018_23 chem_comp_v5 chem_comp_v5_0_2 Y '2018-07-11 11:51:37.958508+00:00' '2018-07-11 11:55:03.966508+00:00'
    # ... abbreviated ...

    """

    def __init__(self, **kwargs):
        self.__startTimestamp = None
        self.__endTimestamp = None
        self.__updateId = 'unset'
        self.__statusFlag = 'N'
        self.__databaseName = 'unset'
        self.__objectName = 'unset'
        self.__tU = TimeUtil()

    def setObject(self, databaseName, objectName):
        """Set the object for current status record.

        Args:
            databaseName (str): database container name
            objectName (str): object name (collection/table) within database

        Returns:
            bool: True for success or False otherwise
        """
        try:
            self.__databaseName = databaseName
            self.__objectName = objectName
            return True
        except Exception as e:
            logger.exception("Failing with %s" % str(e))
        return False

    def setStartTime(self, tS=None, useUtc=True):
        """Set the start time for the current exchange operation.

        Args:
            tS (str, optional): timestamp for the start of the update operation (default=current time)
            useUtc (bool, optional): Report times in UTC

        Returns:
            str: isoformat timestamp or None otherwise
        """
        try:
            self.__startTimestamp = tS if tS else self.__tU.getTimestamp(useUtc=useUtc)
            return self.__startTimestamp
        except Exception as e:
            logger.exception("Failing with %s" % str(e))
        return None

    def setEndTime(self, tS=None, useUtc=True):
        """Set the end time for the current exchange operation.

        Args:
            tS (str, optional): timestamp for the end of the update operation (default=current time)
            useUtc (bool, optional): Report times in UTC

        Returns:
            str: isoformat timestamp or None otherwise
        """
        try:
            self.__endTimestamp = tS if tS else self.__tU.getTimestamp(useUtc=useUtc)
            return self.__endTimestamp
        except Exception as e:
            logger.exception("Failing with %s" % str(e))
        return None

    def setStatus(self, updateId=None, successFlag='Y'):
        """Set the update identifier (yyyy_<week_in_year>) and success flag for the current exchange operation.

        Args:
            updateId (str, optional): Update identifier (default=yyyy_<week_in_year>)
            successFlag (str, optional): 'Y'/'N'

        Returns:
            bool: True for success or False otherwise
        """
        try:
            self.__statusFlag = successFlag
            self.__updateId = updateId if updateId else self.__tU.getCurrentWeekSignature()
            return True
        except Exception as e:
            logger.exception("Failing with %s" % str(e))
        return False

    def getStatus(self, useTimeStrings=False):
        """Get the current data exchange status document.

        Returns:
            dict: Updated list of status records including the appended current record

        """
        try:
            if useTimeStrings:
                sD = {'update_id': self.__updateId,
                      'database_name': self.__databaseName, 'object_name': self.__objectName,
                      'update_status_flag': self.__statusFlag,
                      'update_begin_timestamp': self.__startTimestamp, 'update_end_timestamp': self.__endTimestamp}
            else:
                sD = {'update_id': self.__updateId,
                      'database_name': self.__databaseName, 'object_name': self.__objectName,
                      'update_status_flag': self.__statusFlag,
                      'update_begin_timestamp': self.__tU.getDateTimeObj(self.__startTimestamp), 'update_end_timestamp': self.__tU.getDateTimeObj(self.__endTimestamp)}
            return sD
        except Exception as e:
            logger.exception("Failing with %s" % str(e))
        return {}



##
# File:  Connection.py
# Date:  25-Mar-2018 J. Westbrook
#
# Update:
#   1-Apr-2018 jdw add context methods
#  23-Oct-2018 jdw add section name config access methods and make this a constructor argument
#   5-Dec-2018 jdw pass on exceptions from the context manager __exit__() method
##
"""
Derived class for managing database connection which handles application specific authentication.

"""
__docformat__ = "restructuredtext en"
__author__ = "John Westbrook"
__email__ = "jwest@rcsb.rutgers.edu"
__license__ = "Apache 2.0"

import copy
import logging

from rcsb.db.mongo.ConnectionBase import ConnectionBase

logger = logging.getLogger(__name__)


class Connection(ConnectionBase):

    def __init__(self, cfgOb=None, infoD=None, resourceName=None, sectionName='site_server_info', verbose=False):
        super(Connection, self).__init__(verbose=verbose)
        #
        self.__cfgOb = cfgOb

        #
        if infoD:
            self.setPreferences(infoD)
        #
        if resourceName:
            self.assignResource(resourceName, sectionName)
        #

    def assignResource(self, resourceName=None, sectionName=None):
        #
        defaultPort = 27017
        defaultHost = 'localhost'
        dbServer = 'mongo'

        self._assignResource(resourceName)
        infoD = {}
        if not self.__cfgOb:
            return infoD
        #
        if (resourceName == "EXCHANGE_DB"):
            infoD["DB_NAME"] = self.__cfgOb.get("EXCHANGE_DB_NAME", sectionName=sectionName)
            infoD["DB_HOST"] = self.__cfgOb.get("EXCHANGE_DB_HOST", sectionName=sectionName)
            infoD["DB_SOCKET"] = self.__cfgOb.get("EXCHANGE_DB_SOCKET", default=None, sectionName=sectionName)
            infoD["DB_PORT"] = int(str(self.__cfgOb.get("EXCHANGE_DB_PORT", default=defaultPort, sectionName=sectionName)))
            infoD["DB_USER"] = self.__cfgOb.get("EXCHANGE_DB_USER_NAME", sectionName=sectionName)
            infoD["DB_PW"] = self.__cfgOb.get("EXCHANGE_DB_PASSWORD", sectionName=sectionName)
            infoD["DB_ADMIN_DB_NAME"] = self.__cfgOb.get("EXCHANGE_DB_ADMIN_DB_NAME", default='admin', sectionName=sectionName)
            infoD["DB_WRITE_CONCERN"] = self.__cfgOb.get("EXCHANGE_DB_WRITE_CONCERN", default="majority", sectionName=sectionName)
            infoD["DB_READ_CONCERN"] = self.__cfgOb.get("EXCHANGE_DB_READ_CONCERN", default="majority", sectionName=sectionName)
            infoD["DB_READ_PREFERENCE"] = self.__cfgOb.get("EXCHANGE_DB_READ_PREFERENCE", default="nearest", sectionName=sectionName)
            infoD["DB_WRITE_TO_JOURNAL"] = self.__cfgOb.get("EXCHANGE_DB_WRITE_TO_JOURNAL", default=True, sectionName=sectionName)
        elif (resourceName == "MONGO_DB"):
            infoD["DB_NAME"] = self.__cfgOb.get("MONGO_DB_NAME", sectionName=sectionName)
            infoD["DB_HOST"] = self.__cfgOb.get("MONGO_DB_HOST", default=defaultHost, sectionName=sectionName)
            infoD["DB_SOCKET"] = self.__cfgOb.get("MONGO_DB_SOCKET", default=None, sectionName=sectionName)
            infoD["DB_PORT"] = int(str(self.__cfgOb.get("MONGO_DB_PORT", default=defaultPort, sectionName=sectionName)))
            infoD["DB_USER"] = self.__cfgOb.get("MONGO_DB_USER_NAME", sectionName=sectionName)
            infoD["DB_PW"] = self.__cfgOb.get("MONGO_DB_PASSWORD", sectionName=sectionName)
            infoD["DB_ADMIN_DB_NAME"] = self.__cfgOb.get("MONGO_DB_ADMIN_DB_NAME", default='admin', sectionName=sectionName)
            infoD["DB_WRITE_CONCERN"] = self.__cfgOb.get("MONGO_DB_WRITE_CONCERN", default="majority", sectionName=sectionName)
            infoD["DB_READ_CONCERN"] = self.__cfgOb.get("MONGO_DB_READ_CONCERN", default="majority", sectionName=sectionName)
            infoD["DB_READ_PREFERENCE"] = self.__cfgOb.get("MONGO_DB_READ_PREFERENCE", default="nearest", sectionName=sectionName)
            infoD["DB_WRITE_TO_JOURNAL"] = self.__cfgOb.get("MONGO_DB_WRITE_TO_JOURNAL", default=True, sectionName=sectionName)
        else:
            infoD["DB_NAME"] = self.__cfgOb.get("DB_NAME", sectionName=sectionName)
            infoD["DB_HOST"] = self.__cfgOb.get("DB_HOST", default=defaultHost, sectionName=sectionName)
            infoD["DB_SOCKET"] = self.__cfgOb.get("DB_SOCKET", default=None, sectionName=sectionName)
            infoD["DB_PORT"] = int(str(self.__cfgOb.get("DB_PORT", default=defaultPort, sectionName=sectionName)))
            infoD["DB_USER"] = self.__cfgOb.get("DB_USER_NAME", sectionName=sectionName)
            infoD["DB_PW"] = self.__cfgOb.get("DB_PASSWORD", sectionName=sectionName)
            infoD["DB_ADMIN_DB_NAME"] = self.__cfgOb.get("DB_ADMIN_DB_NAME", default='admin', sectionName=sectionName)
            infoD["DB_WRITE_CONCERN"] = self.__cfgOb.get("DB_WRITE_CONCERN", default="majority", sectionName=sectionName)
            infoD["DB_READ_CONCERN"] = self.__cfgOb.get("DB_READ_CONCERN", default="majority", sectionName=sectionName)
            infoD["DB_READ_PREFERENCE"] = self.__cfgOb.get("DB_READ_PREFERENCE", default="nearest", sectionName=sectionName)
            infoD["DB_WRITE_TO_JOURNAL"] = self.__cfgOb.get("DB_WRITE_TO_JOURNAL", default=True, sectionName=sectionName)
        #
        infoD['DB_SERVER'] = dbServer
        self.setPreferences(infoD)
        #
        return copy.deepcopy(infoD)
        #

    def __enter__(self):
        self.openConnection()
        return self.getClientConnection()

    def __exit__(self, *args):
        if args[0]:
            raise
        return self.closeConnection()



##
# File:    SchemaDocumentHelperBase.py
# Author:  J. Westbrook
# Date:    6-Jun-2018
# Version: 0.001 Initial version
#
# Updates:
#  6-Jun-2018  jdw  separate table and document
# 14-Aug-2018  jdw  generalize document key attribute to attribute list
##
"""
Inject additional information into a schema document definition. Single source of additional document
schema semantic content.

"""
__docformat__ = "restructuredtext en"
__author__ = "John Westbrook"
__email__ = "jwest@rcsb.rutgers.edu"
__license__ = "Apache 2.0"

import logging

logger = logging.getLogger(__name__)


class SchemaDocumentHelperBase(object):
    """ Inject additional information into a schema document definition.

    """

    def __init__(self, **kwargs):
        self._raiseExceptions = kwargs.get('raiseExceptions', False)

    def getCollections(self, schemaName):
        if self._raiseExceptions:
            raise NotImplementedError()
        return []

    def getExcluded(self, collectionName):
        '''  For input collection, return the list of excluded schema identifiers.

        '''
        if self._raiseExceptions:
            raise NotImplementedError()
        return []

    def getIncluded(self, collectionName):
        '''  For input collection, return the list of included schema identifiers.

        '''
        if self._raiseExceptions:
            raise NotImplementedError()
        return []

    def getDocumentKeyAttributeNames(self, collectionName):
        if self._raiseExceptions:
            raise NotImplementedError()
        return []



##
# File:    DictMethodRunner.py
# Author:  J. Westbrook
# Date:    18-Aug-2018
# Version: 0.001 Initial version
#
# Updates:
# 12-Nov-2018 jdw Run block methods after category and attribute methods.
#
##
"""
Manage the invocation of dictionary methods implemented in helper classes.

"""
__docformat__ = "restructuredtext en"
__author__ = "John Westbrook"
__email__ = "jwest@rcsb.rutgers.edu"
__license__ = "Apache 2.0"

import logging
from operator import itemgetter

from mmcif.api.DictionaryApi import DictionaryApi

from rcsb.utils.io.MarshalUtil import MarshalUtil

logger = logging.getLogger(__name__)


class DictMethodRunner(object):
    """ Manage the invocation of dictionary methods implemented in helper classes.

    """

    def __init__(self, dictLocators, methodHelper=None, **kwargs):
        """
        Args:
            dictLocators (list string): dictionary locator list
            methodHelper (class instance, optional): a subclass of of methodoHelperBase.

        """
        self.__dApi = self.__setup(dictLocators, **kwargs)
        self.__methodD = self.__getMethodInfo()
        logger.debug("Method index %r" % self.__methodD.items())
        self.__methodHelper = methodHelper

    def __setup(self, dictLocators, **kwargs):
        """ Return an instance of a dictionary API instance for the input dictionary list
        """
        mU = MarshalUtil()
        containerList = []
        for dictLocator in dictLocators:
            containerList.extend(mU.doImport(dictLocator, format="mmcif-dict"))
        #
        dApi = DictionaryApi(containerList=containerList, consolidate=True, verbose=True)
        return dApi

    def __getMethodInfo(self):
        methodD = {}
        methodIndex = self.__dApi.getMethodIndex()
        for item, mrL in methodIndex.items():
            for mr in mrL:
                mId = mr.getId()
                catName = mr.getCategoryName()
                atName = mr.getAttributeName()
                mType = mr.getType()
                if (catName, atName) not in methodD:
                    methodD[(catName, atName)] = []
                methDef = self.__dApi.getMethod(mId)
                mLang = methDef.getLanguage()
                mCode = methDef.getCode()
                mImplement = methDef.getInline()
                mPriority = methDef.getPriority()
                d = {'METHOD_LANGUAGE': mLang, 'METHOD_IMPLEMENT': mImplement, 'METHOD_TYPE': mType, 'METHOD_CODE': mCode, 'METHOD_PRIORITY': mPriority}
                methodD[(catName, atName)].append(d)
            #
        ##
        logger.debug("Method dictionary %r" % methodD)
        return methodD

    def __invokeAttributeMethod(self, dataContainer, catName, atName, methodName):
        """
        """
        ok = False
        try:
            theMeth = getattr(self.__methodHelper, methodName, None)
            ok = theMeth(dataContainer, catName, atName)
        except Exception as e:
            logger.exception("Failed invoking attribute %s %s method %r with %s" % (catName, atName, methodName, str(e)))
        return ok

    def __invokeCategoryMethod(self, dataContainer, catName, methodName):
        """
        """
        ok = False
        try:
            theMeth = getattr(self.__methodHelper, methodName, None)
            ok = theMeth(dataContainer, catName)
        except Exception as e:
            logger.exception("Failed invoking category %s method %r with %s" % (catName, methodName, str(e)))
        return ok

    def __invokeDatablockMethod(self, dataContainer, blockName, methodName):
        """
        """
        ok = False
        try:
            theMeth = getattr(self.__methodHelper, methodName, None)
            ok = theMeth(dataContainer, blockName)
        except Exception as e:
            logger.exception("Failed invoking block %s method %r with %s" % (blockName, methodName, str(e)))
        return ok

    def apply(self, dataContainer):
        """ Apply datablock, category and attribute dictionary methods on the input data container.
        """
        mTupL = self.__getCategoryMethods(dataContainer)
        logger.debug("Category methods %r" % mTupL)
        for catName, _, methodName, _ in mTupL:
            self.__invokeCategoryMethod(dataContainer, catName, methodName)

        mTupL = self.__getAttributeMethods(dataContainer)
        logger.debug("Attribute methods %r" % mTupL)
        for catName, atName, methodName, _ in mTupL:
            self.__invokeAttributeMethod(dataContainer, catName, atName, methodName)

        mTupL = self.__getDatablockMethods(dataContainer)
        logger.debug("Datablock methods %r" % mTupL)
        for blockName, _, methodName, _ in mTupL:
            self.__invokeDatablockMethod(dataContainer, blockName, methodName)

        return True

    def __getDatablockMethods(self, container, methodCodes=["calculate_with_helper"]):
        mL = []
        try:
            for (dictName, _), mDL in self.__methodD.items():
                for mD in mDL:
                    if mD['METHOD_CODE'].lower() in methodCodes and mD['METHOD_TYPE'].lower() == 'datablock':
                        tL = mD['METHOD_IMPLEMENT'].split('.')
                        methodName = ''.join(tL[1:])
                        mL.append((dictName, None, methodName, mD['METHOD_PRIORITY']))
            mL = sorted(mL, key=itemgetter(3))
            return mL
        except Exception as e:
            logger.exception("Failing dictName %s with %s" % (dictName, str(e)))
        return mL

    def __getCategoryMethods(self, container, methodCodes=["calculate_with_helper"]):
        mL = []
        try:
            for (catName, _), mDL in self.__methodD.items():
                for mD in mDL:
                    if mD['METHOD_CODE'].lower() in methodCodes and mD['METHOD_TYPE'].lower() == 'category':
                        tL = mD['METHOD_IMPLEMENT'].split('.')
                        methodName = ''.join(tL[1:])
                        mL.append((catName, None, methodName, mD['METHOD_PRIORITY']))
            mL = sorted(mL, key=itemgetter(3))
            return mL
        except Exception as e:
            logger.exception("Failing catName %r with %s" % (catName, str(e)))
        return mL

    def __getAttributeMethods(self, container, methodCodes=["calculate_with_helper"]):
        mL = []
        try:
            for (catName, atName), mDL in self.__methodD.items():
                for mD in mDL:
                    if mD['METHOD_CODE'].lower() in methodCodes and mD['METHOD_TYPE'].lower() == 'attribute':
                        tL = mD['METHOD_IMPLEMENT'].split('.')
                        methodName = ''.join(tL[1:])
                        mL.append((catName, atName, methodName, mD['METHOD_PRIORITY']))
            mL = sorted(mL, key=itemgetter(3))
            return mL
        except Exception as e:
            logger.exception("Failing catName %s atName %s with %s" % (catName, atName, str(e)))
        return mL
